# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#          Francesco Bartoli <xbartolone@gmail.com>
#          Sander Schaminee <sander.schaminee@geocat.net>
#          John A Stevenson <jostev@bgs.ac.uk>
#          Colin Blackburn <colb@bgs.ac.uk>
#          Ricardo Garcia Silva <ricardo.garcia.silva@geobeyond.it>
#          Bernhard Mallinger <bernhard.mallinger@eox.at>
#
# Copyright (c) 2026 Tom Kralidis
# Copyright (c) 2025 Francesco Bartoli
# Copyright (c) 2022 John A Stevenson and Colin Blackburn
# Copyright (c) 2023 Ricardo Garcia Silva
# Copyright (c) 2024 Bernhard Mallinger
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from copy import deepcopy
from http import HTTPStatus
import logging
from typing import Tuple
import urllib

from pyproj.exceptions import CRSError
from shapely.errors import ShapelyError
from shapely.wkt import loads as shapely_loads

from pygeoapi import l10n
from pygeoapi.api import evaluate_limit
from pygeoapi.formats import F_COVERAGEJSON, F_HTML, F_JSON, F_JSONLD
from pygeoapi.formatter.base import FormatterSerializationError
from pygeoapi.crs import (create_crs_transform_spec, set_content_crs_header)
from pygeoapi.openapi import get_oas_30_parameters
from pygeoapi.plugin import load_plugin, PLUGINS
from pygeoapi.provider import filter_providers_by_type, get_provider_by_type
from pygeoapi.provider.base import (
    ProviderGenericError, ProviderItemNotFoundError)
from pygeoapi.util import (get_dataset_formatters, get_typed_value,
                           render_j2_template, to_json,
                           filter_dict_by_key_value)

from . import APIRequest, API, validate_datetime, validate_bbox

LOGGER = logging.getLogger(__name__)

CONFORMANCE_CLASSES = [
    'http://www.opengis.net/spec/ogcapi-edr-1/1.0/conf/core'
]


def get_collection_edr_instances(api: API, request: APIRequest,
                                 dataset: str,
                                 instance_id: str | None = None
                                 ) -> Tuple[dict, int, str]:
    """
    Queries collection EDR instances

    :param request: APIRequest instance with query params
    :param dataset: dataset name

    :returns: tuple of headers, status code, content
    """

    data = {
        'instances': [],
        'links': []
    }

    if not request.is_valid(PLUGINS['formatter'].keys()):
        return api.get_format_exception(request)
    headers = request.get_response_headers(api.default_locale,
                                           **api.api_headers)
    collections = filter_dict_by_key_value(api.config['resources'],
                                           'type', 'collection')

    if dataset not in collections.keys():
        msg = 'Collection not found'
        return api.get_exception(
            HTTPStatus.NOT_FOUND, headers, request.format, 'NotFound', msg)

    uri = f'{api.get_collections_url()}/{dataset}'

    LOGGER.debug('Loading provider')
    try:
        p = load_plugin('provider', get_provider_by_type(
            collections[dataset]['providers'], 'edr'))
    except ProviderGenericError as err:
        return api.get_exception(
            err.http_status_code, headers, request.format,
            err.ogc_exception_code, err.message)

    if instance_id is not None:
        try:
            if p.get_instance(instance_id):
                instances = [instance_id]
        except ProviderItemNotFoundError:
            msg = 'Instance not found'
            return api.get_exception(
                HTTPStatus.NOT_FOUND, headers, request.format, 'NotFound', msg)
    else:
        instances = p.get_instances()

    for instance in instances:
        instance_dict = {
            'id': instance,
            'links': [{
                'href': f'{uri}/instances/{instance}?f={F_JSON}',
                'rel': request.get_linkrel(F_JSON),
                'type': 'application/json'
            }, {
                'href': f'{uri}/instances/{instance}?f={F_HTML}',
                'rel': request.get_linkrel(F_HTML),
                'type': 'text/html'
            }, {
                'href': f'{uri}?f={F_HTML}',
                'rel': 'collection',
                'title': collections[dataset]['title'],
                'type': 'text/html'
            }, {
                'href': f'{uri}?f={F_JSON}',
                'rel': 'collection',
                'title': collections[dataset]['title'],
                'type': 'application/json'
            }],
            'data_queries': {}
        }

        for qt in p.get_query_types():
            if qt == 'instances':
                continue

            data_query = {
                'link': {
                    'href': f'{uri}/instances/{instance}/{qt}?f={request.format}',  # noqa
                    'rel': 'data',
                    'title': f'{qt} query'
                }
            }

            if request.format is not None and request.format == 'json':
                data_query['link']['type'] = 'application/vnd.cov+json'

            instance_dict['data_queries'][qt] = data_query

        data['instances'].append(instance_dict)

        if instance_id is not None:
            data = data['instances'][0]
            data.pop('instances', None)
            links_uri = f'{uri}/instances/{instance_id}'
        else:
            links_uri = f'{uri}/instances'

        if instance_id is None:
            data['links'].extend([{
                'href': f'{links_uri}?f={F_JSON}',
                'rel': request.get_linkrel(F_JSON),
                'type': 'application/json'
            }, {
                'href': f'{links_uri}?f={F_HTML}',
                'rel': request.get_linkrel(F_HTML),
                'type': 'text/html'
            }])

    if request.format == F_HTML:  # render
        tpl_config = api.get_dataset_templates(dataset)

        serialized_query_params = ''
        for k, v in request.params.items():
            if k != 'f':
                serialized_query_params += '&'
                serialized_query_params += urllib.parse.quote(k, safe='')
                serialized_query_params += '='
                serialized_query_params += urllib.parse.quote(str(v), safe=',')

        if instance_id is None:
            uri = f'{uri}/instances'
        else:
            uri = f'{uri}/instances/{instance_id}'

        data['query_type'] = 'instances'
        data['query_path'] = uri
        data['title'] = collections[dataset]['title']
        data['description'] = collections[dataset]['description']
        data['keywords'] = collections[dataset]['keywords']
        data['collections_path'] = api.get_collections_url()

        if instance_id is None:
            data['dataset_path'] = data['collections_path'] + '/' + uri.split('/')[-2]  # noqa
            template = 'collections/edr/instances.html'
        else:
            data['dataset_path'] = data['collections_path'] + '/' + uri.split('/')[-3]  # noqa
            template = 'collections/edr/instance.html'

        data['links'] = [{
            'rel': 'collection',
            'title': collections[dataset]['title'],
            'href': f"{data['dataset_path']}?f={F_JSON}",
            'type': 'text/html'
        }, {
            'rel': 'collection',
            'title': collections[dataset]['title'],
            'href': f"{data['dataset_path']}?f={F_HTML}",
            'type': 'application/json'
        }, {
            'type': 'application/json',
            'rel': 'alternate',
            'title': l10n.translate('This document as JSON', request.locale),
            'href': f'{uri}?f={F_JSON}{serialized_query_params}'
        }, {
            'type': 'application/ld+json',
            'rel': 'alternate',
            'title': l10n.translate('This document as JSON-LD', request.locale),  # noqa
            'href': f'{uri}?f={F_JSONLD}{serialized_query_params}'
        }]

        content = render_j2_template(api.tpl_config, tpl_config, template,
                                     data, api.default_locale)
    else:
        content = to_json(data, api.pretty_print)

    return headers, HTTPStatus.OK, content


def get_collection_edr_query(api: API, request: APIRequest,
                             dataset: str, instance: str,
                             query_type: str,
                             location_id: str | None = None
                             ) -> Tuple[dict, int, str]:
    """
    Queries collection EDR

    :param request: APIRequest instance with query params
    :param dataset: dataset name
    :param instance: instance name
    :param query_type: EDR query type
    :param location_id: location id of a /location/<location_id> query

    :returns: tuple of headers, status code, content
    """

    headers = request.get_response_headers(api.default_locale,
                                           **api.api_headers)
    collections = filter_dict_by_key_value(api.config['resources'],
                                           'type', 'collection')

    if dataset not in collections.keys():
        msg = 'Collection not found'
        return api.get_exception(
            HTTPStatus.NOT_FOUND, headers, request.format, 'NotFound', msg)

    LOGGER.debug('Loading provider')
    try:
        provider_def = get_provider_by_type(
            collections[dataset]['providers'], 'edr')
        p = load_plugin('provider', provider_def)
    except ProviderGenericError as err:
        return api.get_exception(
            err.http_status_code, headers, request.format,
            err.ogc_exception_code, err.message)

    if instance is not None and not p.get_instance(instance):
        msg = 'Invalid instance identifier'
        return api.get_exception(
            HTTPStatus.BAD_REQUEST, headers,
            request.format, 'InvalidParameterValue', msg)

    if query_type not in p.get_query_types():
        msg = 'Unsupported query type'
        return api.get_exception(
            HTTPStatus.BAD_REQUEST, headers, request.format,
            'InvalidParameterValue', msg)

    LOGGER.debug('Validating requested format')
    dataset_formatters = get_dataset_formatters(collections[dataset])

    if dataset_formatters:
        LOGGER.debug(f'Dataset formatters: {dataset_formatters}')
        request._format = request._get_format(
            request.get_request_headers(request.headers),
            {v.f: v.mimetype for v in dataset_formatters.values()})

        LOGGER.debug(f'Request format: {request.format}')

    if not request.is_valid(dataset_formatters.keys()):
        return api.get_format_exception(request)

    crs_transform_spec = None
    query_crs_uri = request.params.get('crs')
    if query_crs_uri is not None:
        LOGGER.debug('Processing crs parameter')
        try:
            crs_transform_spec = create_crs_transform_spec(
                provider_def, query_crs_uri
            )
        except (ValueError, CRSError) as err:
            msg = str(err)
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'InvalidParameterValue', msg)
        set_content_crs_header(headers, provider_def, query_crs_uri)

    LOGGER.debug('Processing query parameters')

    LOGGER.debug('Processing datetime parameter')
    datetime_ = request.params.get('datetime')
    try:
        datetime_ = validate_datetime(collections[dataset]['extents'],
                                      datetime_)
    except ValueError as err:
        msg = str(err)
        return api.get_exception(
            HTTPStatus.BAD_REQUEST, headers, request.format,
            'InvalidParameterValue', msg)

    LOGGER.debug('Processing parameter-name parameter')
    parameternames = request.params.get('parameter-name') or []
    if isinstance(parameternames, str):
        parameternames = parameternames.split(',')

    bbox = None
    if query_type in ['cube', 'locations']:
        LOGGER.debug('Processing cube bbox')
        try:
            bbox = validate_bbox(request.params.get('bbox'))
            if not bbox and query_type == 'cube':
                raise ValueError('bbox parameter required by cube queries')
        except ValueError as err:
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'InvalidParameterValue', str(err))

    LOGGER.debug('Processing coords parameter')
    wkt = request.params.get('coords')

    if wkt:
        try:
            wkt = shapely_loads(wkt)
        except ShapelyError:
            msg = 'invalid coords parameter'
            return api.get_exception(
                HTTPStatus.BAD_REQUEST, headers, request.format,
                'InvalidParameterValue', msg)
    elif query_type not in ['cube', 'locations']:
        msg = 'missing coords parameter'
        return api.get_exception(
            HTTPStatus.BAD_REQUEST, headers, request.format,
            'InvalidParameterValue', msg)

    within = within_units = None
    if query_type == 'radius':
        LOGGER.debug('Processing within / within-units parameters')
        within = request.params.get('within')
        within_units = request.params.get('within-units')

    corridor_width = width_units = None
    corridor_height = height_units = None
    if query_type == 'corridor':
        LOGGER.debug('Processing corridor width / height / units parameters')
        corridor_width = request.params.get('corridor-width')
        width_units = request.params.get('width-units')
        corridor_height = request.params.get('corridor-height')
        height_units = request.params.get('height-units')

    LOGGER.debug('Processing z parameter')
    try:
        z = get_typed_value(request.params.get('z'))
    except TypeError:
        z = None

    if parameternames and not any((fld in parameternames)
                                  for fld in p.get_fields().keys()):
        msg = 'Invalid parameter-name'
        return api.get_exception(
            HTTPStatus.BAD_REQUEST, headers, request.format,
            'InvalidParameterValue', msg)

    LOGGER.debug('Processing limit parameter')
    if api.config['server'].get('limit') is not None:
        msg = ('server.limit is no longer supported! '
               'Please use limits at the server or collection '
               'level (RFC5)')
        LOGGER.warning(msg)
    try:
        limit = evaluate_limit(request.params.get('limit'),
                               api.config['server'].get('limits', {}),
                               collections[dataset].get('limits', {}))
    except ValueError as err:
        return api.get_exception(
            HTTPStatus.BAD_REQUEST, headers, request.format,
            'InvalidParameterValue', str(err))

    query_args = dict(
        query_type=query_type,
        instance=instance,
        format_=request.format,
        datetime_=datetime_,
        select_properties=parameternames,
        wkt=wkt,
        z=z,
        bbox=bbox,
        within=within,
        within_units=within_units,
        corridor_width=corridor_width,
        width_units=width_units,
        corridor_height=corridor_height,
        height_units=height_units,
        limit=limit,
        location_id=location_id,
        crs_transform_spec=crs_transform_spec
    )

    try:
        data = p.query(**query_args)
    except ProviderGenericError as err:
        return api.get_exception(
            err.http_status_code, headers, request.format,
            err.ogc_exception_code, err.message)

    if request.format == F_HTML:  # render
        tpl_config = api.get_dataset_templates(dataset)

        uri = f'{api.get_collections_url()}/{dataset}/{query_type}'
        serialized_query_params = ''
        for k, v in request.params.items():
            if k != 'f':
                serialized_query_params += '&'
                serialized_query_params += urllib.parse.quote(k, safe='')
                serialized_query_params += '='
                serialized_query_params += urllib.parse.quote(str(v), safe=',')

        data['query_type'] = query_type.capitalize()
        data['query_path'] = uri
        data['dataset_path'] = '/'.join(uri.split('/')[:-1])
        data['collections_path'] = api.get_collections_url()

        data['links'] = [{
            'rel': 'collection',
            'title': collections[dataset]['title'],
            'href': data['dataset_path']
        }, {
            'type': 'application/prs.coverage+json',
            'rel': request.get_linkrel(F_COVERAGEJSON),
            'title': l10n.translate('This document as CoverageJSON', request.locale),  # noqa
            'href': f'{uri}?f={F_COVERAGEJSON}{serialized_query_params}'
        }, {
            'type': 'application/ld+json',
            'rel': 'alternate',
            'title': l10n.translate('This document as JSON-LD', request.locale),  # noqa
            'href': f'{uri}?f={F_JSONLD}{serialized_query_params}'
        }]

        content = render_j2_template(api.tpl_config, tpl_config,
                                     'collections/edr/query.html', data,
                                     api.default_locale)
    elif request.format in [df.f for df in dataset_formatters.values()]:
        formatter = [v for v in dataset_formatters.values() if
                     v.f == request.format][0]

        try:
            content = formatter.write(
                data=data,
                options={
                    'provider_def': get_provider_by_type(
                        collections[dataset]['providers'],
                        'edr')
                }
            )
        except FormatterSerializationError:
            msg = 'Error serializing output'
            return api.get_exception(
                HTTPStatus.INTERNAL_SERVER_ERROR, headers, request.format,
                'NoApplicableCode', msg)

        if formatter.attachment:
            filename = f'{dataset}.{formatter.extension}'
            cd = f'attachment; filename="{filename}"'
            headers['Content-Disposition'] = cd

    else:
        headers['Content-Type'] = 'application/vnd.cov+json'
        content = to_json(data, api.pretty_print)

    return headers, HTTPStatus.OK, content


def get_oas_30(cfg: dict, locale: str) -> tuple[list[dict[str, str]], dict[str, dict]]:  # noqa
    """
    Get OpenAPI fragments

    :param cfg: `dict` of configuration
    :param locale: `str` of locale

    :returns: `tuple` of `list` of tag objects, and `dict` of path objects
    """

    from pygeoapi.openapi import OPENAPI_YAML, get_visible_collections

    LOGGER.debug('setting up edr endpoints')

    paths = {}

    collections = filter_dict_by_key_value(cfg['resources'],
                                           'type', 'collection')

    for k, v in get_visible_collections(cfg).items():
        edr_extension = filter_providers_by_type(
            collections[k]['providers'], 'edr')

        description = l10n.translate(v['description'], locale)

        if edr_extension:
            collection_name_path = f'/collections/{k}'

            ep = load_plugin('provider', edr_extension)

            edr_query_endpoints = []

            for qt in [qt for qt in ep.get_query_types() if qt != 'locations']:
                edr_query_endpoints.append({
                    'path': f'{collection_name_path}/{qt}',
                    'qt': qt,
                    'op_id': f'query{qt.capitalize()}{k.capitalize()}'
                })
                if 'instances' in ep.get_query_types() and qt != 'instances':
                    edr_query_endpoints.append({
                        'path': f'{collection_name_path}/instances/{{instanceId}}/{qt}',  # noqa
                        'qt': qt,
                        'op_id': f'query{qt.capitalize()}Instance{k.capitalize()}'  # noqa
                    })

            for eqe in edr_query_endpoints:
                if eqe['qt'] == 'cube':
                    spatial_parameter = {
                        'description': 'Only features that have a geometry that intersects the bounding box are selected.The bounding box is provided as four or six numbers, depending on whether the coordinate reference system includes a vertical axis (height or depth).',  # noqa
                        'explode': False,
                        'in': 'query',
                        'name': 'bbox',
                        'required': True,
                        'schema': {
                            'items': {
                                'type': 'number'
                            },
                            'maxItems': 6,
                            'minItems': 4,
                            'type': 'array'
                        },
                        'style': 'form'
                    }
                else:
                    spatial_parameter = {
                        '$ref': f"{OPENAPI_YAML['oaedr']}/parameters/{eqe['qt']}Coords.yaml"  # noqa
                    }

                dataset_formatters = get_dataset_formatters(v)
                coll_f_parameter = deepcopy(get_oas_30_parameters(cfg, locale))['f']  # noqa
                for key, value in dataset_formatters.items():
                    coll_f_parameter['schema']['enum'].append(value.f)

                paths[eqe['path']] = {
                    'get': {
                        'summary': f"query {description} by {eqe['qt']}",
                        'description': description,
                        'tags': [k],
                        'operationId': eqe['op_id'],
                        'parameters': [
                            spatial_parameter,
                            {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/parameters/datetime"},  # noqa
                            {'$ref': f"{OPENAPI_YAML['oaedr']}/parameters/parameter-name.yaml"},  # noqa
                            {'$ref': f"{OPENAPI_YAML['oaedr']}/parameters/z.yaml"},  # noqa
                            {'$ref': '#/components/parameters/crs'},
                            coll_f_parameter,
                        ],
                        'responses': {
                            '200': {
                                'description': 'Response',
                                'content': {
                                    'application/prs.coverage+json': {
                                        'schema': {
                                            '$ref': f"{OPENAPI_YAML['oaedr']}/schemas/coverageJSON.yaml"  # noqa
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                if 'instanceId' in eqe['path']:
                    paths[eqe['path']]['get']['parameters'].insert(0, {
                        '$ref': f"{OPENAPI_YAML['oaedr']}/parameters/instanceId.yaml"}  # noqa
                    )

            if 'instances' in ep.get_query_types():
                paths[f'{collection_name_path}/instances'] = {
                    'get': {
                        'summary': f"Get pre-defined instances of {description}",  # noqa
                        'description': description,
                        'tags': [k],
                        'operationId': f'getInstances{k.capitalize()}',
                        'parameters': [
                            {'$ref': '#/components/parameters/f'}
                        ],
                        'responses': {
                            '200': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/Features"},  # noqa
                            '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                            '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"}  # noqa
                        }
                    }
                }
                paths[f'{collection_name_path}/instances/{{instanceId}}'] = {
                    'get': {
                        'summary': f"Get {description} instance",
                        'description': description,
                        'tags': [k],
                        'operationId': f'getInstance{k.capitalize()}',
                        'parameters': [
                            {'$ref': f"{OPENAPI_YAML['oaedr']}/parameters/instanceId.yaml"},  # noqa
                            {'$ref': '#/components/parameters/f'}
                        ],
                        'responses': {
                            '200': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/Features"},  # noqa
                        }
                    }
                }

            if 'locations' in ep.get_query_types():
                paths[f'{collection_name_path}/locations'] = {
                    'get': {
                        'summary': f"Get pre-defined locations of {description}",  # noqa
                        'description': description,
                        'tags': [k],
                        'operationId': f'getLocations{k.capitalize()}',
                        'parameters': [
                            {'$ref': f"{OPENAPI_YAML['oaedr']}/parameters/bbox.yaml"},  # noqa
                            {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/parameters/datetime"},  # noqa
                            {'$ref': '#/components/parameters/f'}
                        ],
                        'responses': {
                            '200': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/Features"},  # noqa
                            '400': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/InvalidParameter"},  # noqa
                            '500': {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/responses/ServerError"}  # noqa
                        }
                    }
                }
                paths[f'{collection_name_path}/locations/{{locId}}'] = {
                    'get': {
                        'summary': f"query {description} by location",
                        'description': description,
                        'tags': [k],
                        'operationId': f'getLocation{k.capitalize()}',
                        'parameters': [
                            {'$ref': f"{OPENAPI_YAML['oaedr']}/parameters/locationId.yaml"},  # noqa
                            {'$ref': f"{OPENAPI_YAML['oapif-1']}#/components/parameters/datetime"},  # noqa
                            {'$ref': f"{OPENAPI_YAML['oaedr']}/parameters/parameter-name.yaml"},  # noqa
                            {'$ref': '#/components/parameters/crs'},
                            {'$ref': '#/components/parameters/f'}
                        ],
                        'responses': {
                            '200': {
                                'description': 'Response',
                                'content': {
                                    'application/prs.coverage+json': {
                                        'schema': {
                                            '$ref': f"{OPENAPI_YAML['oaedr']}/schemas/coverageJSON.yaml"  # noqa
                                        }
                                    }
                                }
                            }
                        }
                    }
                }

    return [{'name': 'edr'}], {'paths': paths}



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2022 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging

from osgeo import osr
import mapscript
from mapscript import MapServerError

from pygeoapi.provider.base import (BaseProvider, ProviderConnectionError,
                                    ProviderQueryError)
from pygeoapi.util import str2bool

LOGGER = logging.getLogger(__name__)

IMAGE_FORMATS = {
    'png': 'AGG/PNG',  # 24-bit by default
    'png8': 'AGG/PNG8',
    'jpeg': 'AGG/JPEG'
}


class MapScriptProvider(BaseProvider):
    """MapScript map provider (https://mapserver.org/mapscript)"""

    def __init__(self, provider_def):
        """
        Initialize object

        :param provider_def: provider definition

        :returns: pygeoapi.provider.mapscript_.MapScriptProvider
        """

        BaseProvider.__init__(self, provider_def)

        self.crs_list = []
        self.styles = []
        self.default_format = 'png'

        LOGGER.debug(f'MapScript version: {mapscript.MS_VERSION}')

        try:
            LOGGER.debug('Creating new mapObj and layerObj')
            self._map = mapscript.mapObj()
            self._layer = mapscript.layerObj(self._map)
            self._layer.status = mapscript.MS_ON

            file_extension = self.data.split('.')[-1]

            if str2bool(self.options.get('tileindex', False)):
                LOGGER.debug('Setting tileindex')
                self._layer.tileindex = self.data
            else:
                if file_extension in ['shp', 'tif']:
                    LOGGER.debug('Setting built-in MapServer driver')
                    self._layer.data = self.data
                else:
                    LOGGER.debug('Setting OGR driver')
                    self._layer.setConnectionType(mapscript.MS_OGR, 'OGR')
                    self._layer.connection = self.data

            self._layer.type = getattr(mapscript, self.options['type'])

            try:
                self.crs = int(self.options['projection'])
            except KeyError:
                self.crs = 4326

            self._layer.setProjection(self._epsg2projstring(self.crs))

            LOGGER.debug(f'Layer projection: {self._layer.getProjection()}')

            if 'style' in self.options:
                if self.options['style'].endswith(('xml', 'sld')):
                    LOGGER.debug('Setting SLD')
                    with open(self.options['style']) as fh:
                        self._layer.applySLD(fh.read(), self.options['layer'])
                elif self.options['style'].endswith('inc'):
                    LOGGER.debug('Setting MapServer class file')
                    cls = mapscript.classObj(self._layer)
                    with open(self.options['style']) as fh:
                        cls.updateFromString(fh.read())
            else:
                if self.options['type'] != 'MS_LAYER_RASTER':
                    LOGGER.debug('No vector styling found, setting default')
                    cls = mapscript.classObj()
                    cls_def = 'CLASS NAME "default" STYLE COLOR 0 0 0 END END'
                    cls.updateFromString(cls_def)
                    self._layer.insertClass(cls)

            if self.options['type'] == 'MS_LAYER_RASTER':
                self._layer.addProcessing('SCALE=AUTO')

        except MapServerError as err:
            LOGGER.warning(err)
            raise ProviderConnectionError('Cannot connect to map service')

    def query(self, style=None, bbox=[], width=500, height=300, crs='CRS84',
              datetime_=None, format_='png', transparent=True, **kwargs):
        """
        Generate map

        :param style: style name (default is `None`)
        :param bbox: bounding box [minx,miny,maxx,maxy]
        :param width: width of output image (in pixels)
        :param height: height of output image (in pixels)
        :param datetime_: temporal (datestamp or extent)
        :param crs: coordinate reference system identifier
        :param format_: Output format (default is `png`)
        :param transparent: Apply transparency to map (default is `True`)

        :returns: `bytes` of map image
        """

        try:
            image_obj_format = IMAGE_FORMATS[format_]
        except KeyError:
            LOGGER.error(f'Bad output format: {image_obj_format}')
            raise ProviderQueryError('Bad image format')

        LOGGER.debug('Setting output map CRS')
        try:
            if crs not in ['CRS84', 4326]:
                LOGGER.debug('Reprojecting coordinates')
                prj_dst_text = self._epsg2projstring(int(crs.split("/")[-1]))

                prj_src = mapscript.projectionObj(self._layer.getProjection())
                prj_dst = mapscript.projectionObj(prj_dst_text)

                rect = mapscript.rectObj(*bbox)
                _ = rect.project(prj_src, prj_dst)

                map_bbox = [rect.minx, rect.miny, rect.maxx, rect.maxy]
                map_crs = prj_dst_text
                self._map.units = mapscript.MS_METERS

            else:
                map_bbox = bbox
                map_crs = self._epsg2projstring(4326)
                self._map.units = mapscript.MS_DD

        except MapServerError as err:
            LOGGER.error(err)
            raise ProviderQueryError('bad projection')

        if datetime_ is not None:
            if self.time_field is None:
                LOGGER.debug('collection is not time enabled')
            else:
                fe = f'{self.time_field} = "{datetime_}"'
                LOGGER.debug(f'Setting temporal filter: {fe}')
                self._layer.setFilter(fe)

        LOGGER.debug('Setting output image properties')
        fmt = mapscript.outputFormatObj(image_obj_format)
        if transparent:
            fmt.transparent = mapscript.MS_ON
        else:
            fmt.transparent = mapscript.MS_OFF

        self._map.setOutputFormat(fmt)
        self._map.setExtent(*map_bbox)
        self._map.setSize(width, height)

        self._map.setProjection(map_crs)
        self._map.setConfigOption('MS_NONSQUARE', 'yes')

        LOGGER.debug(f'Mapfile: {self._map.convertToString()}')
        try:
            img = self._map.draw()
        except MapServerError as err:
            LOGGER.debug(err)
            raise ProviderQueryError(err)

        return img.getBytes()

    def _epsg2projstring(self, epsg_code):
        """
        Helper function to derive a proj string from an EPSG code

        :param epsg_code: `int` of EPSG code

        :returns: `str` of PROJ string/syntax
        """

        prj = osr.SpatialReference()
        prj.ImportFromEPSG(epsg_code)

        return prj.ExportToProj4().strip()

    def __repr__(self):
        return f'<MapScriptProvider> {self.data}'



# =================================================================
#
# Authors: Andreas Kosubek <andreas.kosubek@ama.gv.at>
# Authors: Moritz Langer <moritz.b.langer@gmail.com>
#
# Copyright (c) 2023 Andreas Kosubek
# Copyright (c) 2024 Moritz Langer
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import importlib
import json
import logging
import os
import threading
from typing import Optional

import oracledb
import pyproj

from pygeoapi.crs import get_srid
from pygeoapi.provider.base import (
    BaseProvider,
    ProviderConnectionError,
    ProviderGenericError,
    ProviderInvalidQueryError,
    ProviderItemNotFoundError,
    ProviderQueryError,
)

LOGGER = logging.getLogger(__name__)


class DatabaseConnection:
    """Database connection class to be used as 'with' statement.
    The class returns a connection object.
    """
    pool = None  # Class-level connection pool
    lock = threading.Lock()

    @classmethod
    def create_pool(cls, conn_dict, oracle_pool_min, oracle_pool_max):
        """Initialize the connection pool for the class
           Lock is implemented before function call at __init__"""
        dsn = cls._make_dsn(conn_dict)

        connect_kwargs = {
            'dsn': dsn,
            'min': oracle_pool_min,
            'max': oracle_pool_max,
            'increment': 1
        }

        # Create the pool
        if conn_dict.get("external_auth") == "wallet":
            # If Auth is via Wallet you need to save a wallet under
            # the directory returned by this bash command if apache is used
            # cat /etc/passwd |grep apache
            # except another directory is specified in the sqlnet.ora file
            LOGGER.debug("Connection pool from wallet.")
            connect_kwargs["externalauth"] = True
            connect_kwargs["homogeneous"] = False

        else:
            LOGGER.debug("Connection pool from user and password.")
            connect_kwargs["user"] = conn_dict["user"]
            connect_kwargs["password"] = conn_dict["password"]

        p = oracledb.create_pool(**connect_kwargs)
        LOGGER.debug("Connection pool created successfully")

        return p

    def __init__(self, conn_dic, table, properties=[], context="query"):
        """
        OracleProvider Class constructor

        :param conn_dic: dictionary with connection parameters
            service_name - the service name of the database instance
            sid - sid of the database instance
            tns_name - name of the tnsnames.ora entry
            external_auth - External authentication e.g. wallet
            tns_admin - path to tnsnames.ora configuration file
            user - user name used to authenticate
            password - password used to authenticate
            host - database host address
            port - connection port number
             (defaults to 1521 if not provided)
            init_oracle_client -

        :param table: table name containing the data. This variable is used to
                assemble column information
        :param properties: User-specified subset of column names to expose
        :param context: query or hits, if query then it will determine
                table column otherwise will not do it
        :returns: DatabaseConnection
        """

        self.conn_dict = conn_dic
        self.table = table
        self.context = context
        self.columns = (
            None  # Comma sepparated string with column names (for SQL query)
        )
        self.properties = [item.lower() for item in properties]
        self.fields = {}  # Dict of columns. Key is col name, value is type
        oracle_pool_min = int(os.environ.get('ORACLE_POOL_MIN', 0))
        oracle_pool_max = int(os.environ.get('ORACLE_POOL_MAX', 0))
        # Initialize the connection pool if it hasn't been initialized
        if oracle_pool_min and oracle_pool_max:
            LOGGER.debug("Found environment variables for session pooling:")
            LOGGER.debug(f"ORACLE_POOL_MIN: {oracle_pool_min}")
            LOGGER.debug(f"ORACLE_POOL_MAX: {oracle_pool_max}")
            with DatabaseConnection.lock:
                if DatabaseConnection.pool is None:
                    LOGGER.debug(f"self.conn_dict contains {self.conn_dict}")
                    DatabaseConnection.pool = DatabaseConnection.create_pool(
                        self.conn_dict, oracle_pool_min, oracle_pool_max
                    )
                    LOGGER.debug(
                        "Initialized connection pool with "
                        f"{DatabaseConnection.pool.max} connections"
                    )

    @staticmethod
    def _make_dsn(conn_dict):
        if conn_dict.get("init_oracle_client", False):
            oracledb.init_oracle_client()

        # Connect with tnsnames.ora entry and Login with Oracle Wallet
        if conn_dict.get("external_auth") == "wallet":
            LOGGER.debug(
                "Oracle connect with tnsnames.ora entry \
                and login with Oracle Wallet"
            )
            if "tns_name" not in conn_dict:
                raise ProviderConnectionError(
                    "tns_name must be set for external authentication!"
                )

            dsn = conn_dict["tns_name"]

        # Connect with SERVICE_NAME
        if "service_name" in conn_dict:
            LOGGER.debug(
                f"Oracle connect with service_name: \
                    {conn_dict['service_name']}"
            )

            if "host" not in conn_dict:
                raise ProviderConnectionError(
                    "Host must be set for connection with service_name!"
                )

            dsn = oracledb.makedsn(
                conn_dict["host"],
                conn_dict.get("port", 1521),
                service_name=conn_dict["service_name"],
            )

        # Connect with SID
        elif "sid" in conn_dict:
            LOGGER.debug(
                f"Oracle connect with sid: {conn_dict['sid']}"
            )

            if "host" not in conn_dict:
                raise ProviderConnectionError(
                    "Host must be set for connection with sid!"
                )

            dsn = oracledb.makedsn(
                conn_dict["host"],
                conn_dict.get("port", 1521),
                sid=conn_dict["sid"],
            )

        # Connect with tnsnames.ora entry
        elif "tns_name" in conn_dict:
            LOGGER.debug(
                f"Oracle connect with tns_name: \
                    {conn_dict['tns_name']}"
            )
            dsn = conn_dict["tns_name"]

        else:
            raise ProviderConnectionError(
                "One of service_name, sid or tns_name must be specified!"
            )

        LOGGER.debug(f"Oracle DSN string: {dsn}")

        return dsn

    def __enter__(self):
        """Acquires a connection from the pool."""
        try:
            if DatabaseConnection.pool:
                self.conn = DatabaseConnection.pool.acquire()
                LOGGER.debug("Connection acquired from pool .")
                LOGGER.debug(f"Connection from pool is {self.conn}.")
            else:
                dsn = self._make_dsn(self.conn_dict)
                LOGGER.debug(f"Created dsn for single connection with params: {dsn}")  # noqa
                # Connect with tnsnames.ora entry and Login with Oracle Wallet  # noqa
                if self.conn_dict.get("external_auth") == "wallet":
                    self.conn = oracledb.connect(externalauth=True, dsn=dsn)

                # Connect with tnsnames.ora entry,
                # TNS_ADMIN is set via configuration
                if "tns_admin" in self.conn_dict:
                    self.conn = oracledb.connect(
                        user=self.conn_dict["user"],
                        password=self.conn_dict["password"],
                        dsn=dsn,
                        config_dir=self.conn_dict["tns_admin"],
                    )

                # Connect with user / password via dsn string
                # When dsn is a TNS name, the environment variable TNS_ADMIN must   # noqa
                # be set (Path to tnsnames.ora file)
                else:
                    self.conn = oracledb.connect(
                        user=self.conn_dict["user"],
                        password=self.conn_dict["password"],
                        dsn=dsn,
                    )

        except oracledb.DatabaseError as e:
            if DatabaseConnection.pool:
                LOGGER.error("Couldn't acquire a connection from the pool.")
                LOGGER.error(e)
            else:
                LOGGER.error(
                    f"Couldn't connect to Oracle using:{str(self.conn_dict)}"
                )
                LOGGER.error(e)
            raise ProviderConnectionError(e)

        # Check if table name has schema/owner inside
        # If not, current user is set
        table_parts = self.table.split(".")
        if len(table_parts) == 2:
            schema = table_parts[0]
            table = table_parts[1]
        else:
            schema = self.conn_dict["user"]
            table = self.table

        LOGGER.debug("Schema: " + schema)
        LOGGER.debug("Table: " + table)

        if self.context == "query":
            columns = dict(self._get_table_columns(schema, table))

            # Populate dictionary for columns with column type
            # NOTE: we want all columns available here because they are
            #       used for filtering in the where clause, not only
            #       the ones that are returned to the client.
            for k, v in columns.items():
                self.fields[k.lower()] = {"type": v}

            filtered_columns = set(self.fields)
            if self.properties:
                filtered_columns &= {k.lower() for k in self.properties}

            # fields which are part of the output
            self.filtered_fields = {
                k: v for k, v in self.fields.items() if k in filtered_columns
            }

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Releases the connection back to the pool.
        """
        try:
            if DatabaseConnection.pool:
                DatabaseConnection.pool.release(self.conn)
                LOGGER.debug("Connection released back to pool.")
            else:
                self.conn.close()
                LOGGER.debug("Single Connection closed")
        except oracledb.DatabaseError as e:
            LOGGER.error("Error closing the connection.")
            LOGGER.error(e)

    def _get_table_columns(self, schema, table):
        """
        Returns an array with all column names and data types
        from Oracle table ALL_TAB_COLUMNS.
        Lookup for public and private synonyms.
        Throws ProviderGenericError when table not exist or accessible.
        """

        sql = """
              SELECT COUNT(1)
                FROM all_objects
               WHERE object_type IN ('VIEW','TABLE','MATERIALIZED VIEW')
                 AND object_name = UPPER(:table_name)
                 AND owner = UPPER(:owner)
              """
        with self.conn.cursor() as cur:
            cur.execute(sql, {"table_name": table, "owner": schema})
            result = cur.fetchone()

        if result[0] == 0:
            sql = """
                  SELECT COUNT(1)
                    FROM all_synonyms
                   WHERE synonym_name = UPPER(:table_name)
                     AND owner = UPPER(:owner)
                  """
            with self.conn.cursor() as cur:
                cur.execute(sql, {"table_name": table, "owner": schema})
                result = cur.fetchone()

            if result[0] == 0:
                sql = """
                      SELECT COUNT(1)
                        FROM all_synonyms
                       WHERE synonym_name = UPPER(:table_name)
                         AND owner = 'PUBLIC'
                      """
                with self.conn.cursor() as cur:
                    cur.execute(sql, {"table_name": table})
                    result = cur.fetchone()

                if result[0] == 0:
                    raise ProviderGenericError(
                        f"Table {schema}.{table} not found!"
                    )

                else:
                    schema = "PUBLIC"

            sql = """
                  SELECT table_owner, table_name
                    FROM all_synonyms
                   WHERE synonym_name = UPPER(:table_name)
                     AND owner = UPPER(:owner)
                  """
            with self.conn.cursor() as cur:
                cur.execute(sql, {"table_name": table, "owner": schema})
                result = cur.fetchone()

            schema = result[0]
            table = result[1]

        # Get table column names and types, excluding geometry
        query_cols = """
                     SELECT column_name, data_type
                       FROM all_tab_columns
                      WHERE table_name = UPPER(:table_name)
                        AND owner = UPPER(:owner)
                        AND data_type != 'SDO_GEOMETRY'
                     """
        with self.conn.cursor() as cur:
            cur.execute(query_cols, {"table_name": table, "owner": schema})
            result = cur.fetchall()

        return result


class OracleProvider(BaseProvider):
    def __init__(self, provider_def):
        """
        OracleProvider Class constructor

        :param provider_def: provider definitions from yml pygeoapi-config.
                             data,id_field, name set in parent class
                             data contains the connection information
                             for class DatabaseCursor

        :returns: pygeoapi.provider.base.OracleProvider
        """

        super().__init__(provider_def)

        # Table properties
        self.table = provider_def["table"]
        self.conn_dic = provider_def["data"]
        self.geom = provider_def["geom_field"]
        self.properties = [item.lower() for item in self.properties]
        self.mandatory_properties = provider_def.get("mandatory_properties")
        self.extra_properties = provider_def.get("extra_properties", [])

        # SQL manipulator properties
        self.sql_manipulator = provider_def.get("sql_manipulator")
        self.sql_manipulator_options = provider_def.get(
            "sql_manipulator_options"
        )

        # SDO properties
        self.sdo_param = provider_def.get("sdo_param")
        self.sdo_operator = provider_def.get("sdo_operator", "sdo_filter")

        LOGGER.debug("Setting Oracle properties:")
        LOGGER.debug(f"Name:{self.name}")
        LOGGER.debug(f"ID_field:{self.id_field}")
        LOGGER.debug(f"Table:{self.table}")
        LOGGER.debug(f"sdo_param: {self.sdo_param}")
        LOGGER.debug(f"sdo_operator: {self.sdo_operator}")
        LOGGER.debug(f"storage_crs {self.storage_crs}")

        # TODO See Issue #1393
        # LOGGER.debug(f"default_crs: {self.default_crs}")

        self.get_fields()

    def get_fields(self):
        """
        Get fields from Oracle table (columns are field)

        :returns: dict of fields
        """
        LOGGER.debug("Get available fields/properties")

        if not self._fields:
            with DatabaseConnection(
                self.conn_dic, self.table, properties=self.properties
            ) as db:
                self._fields = db.fields
        return self._fields

    def _get_where_clauses(
        self,
        properties,
        bbox,
        bbox_crs,
        sdo_param=None,
        sdo_operator="sdo_filter",
    ):
        """
        Generarates WHERE conditions to be implemented in query.
        Private method mainly associated with query method
        :param properties: list of tuples (name, value)
        :param bbox: bounding box [minx,miny,maxx,maxy]

        :returns: Dictionary with sql where clause and bind variables
        """
        LOGGER.debug("Get where clause with bind variables as dictionary")

        where_dict = {"clause": "", "properties": {}}

        where_conditions = []

        if properties:
            prop_clauses = [f"{key} = :{key}" for key, value in properties]
            where_conditions += prop_clauses
            where_dict["properties"] = dict(properties)

        if bbox:
            bbox_dict = {"clause": "", "properties": {}}

            if sdo_operator == "sdo_relate":
                if not sdo_param:
                    sdo_param = "mask=anyinteract"

                bbox_dict["properties"] = {
                    "srid": get_srid(bbox_crs),
                    "minx": bbox[0],
                    "miny": bbox[1],
                    "maxx": bbox[2],
                    "maxy": bbox[3],
                    "sdo_param": sdo_param,
                }

                bbox_query = f"""
                sdo_relate({self.geom},
                           mdsys.sdo_geometry(2003,
                                              :srid,
                                              NULL,
                                              mdsys.sdo_elem_info_array(
                                                    1,
                                                    1003,
                                                    3
                                              ),
                                              mdsys.sdo_ordinate_array(
                                                    :minx,
                                                    :miny,
                                                    :maxx,
                                                    :maxy
                                              )
                           ),
                           :sdo_param
                ) = 'TRUE'
                """

            else:
                bbox_dict["properties"] = {
                    "srid": get_srid(bbox_crs),
                    "minx": bbox[0],
                    "miny": bbox[1],
                    "maxx": bbox[2],
                    "maxy": bbox[3],
                    "sdo_param": sdo_param,
                }

                bbox_query = f"""
                sdo_filter({self.geom},
                           mdsys.sdo_geometry(2003,
                                              :srid,
                                              NULL,
                                              mdsys.sdo_elem_info_array(
                                                    1,
                                                    1003,
                                                    3
                                              ),
                                              mdsys.sdo_ordinate_array(
                                                    :minx,
                                                    :miny,
                                                    :maxx,
                                                    :maxy
                                              )
                           ),
                           :sdo_param
                ) = 'TRUE'
                """

            bbox_dict["clause"] = bbox_query

            where_conditions.append(bbox_dict["clause"])
            where_dict["properties"].update(bbox_dict["properties"])

        if where_conditions:
            where_dict["clause"] = f" WHERE {' AND '.join(where_conditions)}"

        LOGGER.debug(where_dict)

        return where_dict

    def _get_orderby(self, sortby):
        """
        Private function: Get ORDER BY clause

        :param sortby: list of dicts (property, order)

        :returns: STA $orderby string
        """
        sort_map = {"+": "ASC", "-": "DESC"}
        ret = [
            f"{sort['property']} {sort_map[sort['order']]}" for sort in sortby
        ]

        return f"ORDER BY {','.join(ret)}"

    def _get_extra_columns_expression(self):
        """Returns part of SELECT clause for extra properties"""
        return "".join(
            f", {e_prop}" for e_prop in self.extra_properties
        )

    def _output_type_handler(
        self, cursor, name, default_type, size, precision, scale
    ):
        """
        Output type handler for Oracle LOB datatypes
        """
        if default_type == oracledb.DB_TYPE_CLOB:
            return cursor.var(
                oracledb.DB_TYPE_LONG, arraysize=cursor.arraysize
            )
        if default_type == oracledb.DB_TYPE_BLOB:
            return cursor.var(
                oracledb.DB_TYPE_LONG_RAW, arraysize=cursor.arraysize
            )

    def _process_query_with_sql_manipulator_sup(
        self, db, sql_query, bind_variables, extra_params, **query_args
    ):
        """
        Apply the SQL manipulation plugin to process the SQL query.

        :param db: Database connection instance
        :param sql_query: The SQL query to process
        :param bind_variables: Query bind variables
        :param extra_params: Additional parameters for manipulation
        :param query_args: Other dynamic arguments required for processing
        :return: Processed SQL query and bind variables
        """
        if self.sql_manipulator:
            LOGGER.debug(f"sql_manipulator: {self.sql_manipulator}")
            manipulation_class = _class_factory(self.sql_manipulator)

            # Pass all arguments to the process_query method
            sql_query, bind_variables = manipulation_class.process_query(
                db=db,
                sql_query=sql_query,
                bind_variables=bind_variables,
                sql_manipulator_options=self.sql_manipulator_options,
                **query_args,
                extra_params=extra_params,
            )

        for placeholder in ["#HINTS#", "#JOIN#", "#WHERE#"]:
            sql_query = sql_query.replace(placeholder, "")

        return sql_query, bind_variables

    def query(
        self,
        offset=0,
        limit=10,
        resulttype="results",
        bbox=None,
        datetime_=None,
        properties=[],
        sortby=[],
        skip_geometry=False,
        select_properties=[],
        crs_transform_spec=None,
        q=None,
        language=None,
        filterq=None,
        **kwargs,
    ):
        """
        Query Oracle for all the content.

        :param offset: starting record to return (default 0)
        :param limit: number of records to return (default 10)
        :param resulttype: return results or hit limit (default results)
        :param bbox: bounding box [minx,miny,maxx,maxy]
        :param datetime_: temporal (datestamp or extent)
        :param properties: list of tuples (name, value)
        :param sortby: list of dicts (property, order)
        :param select_properties: list of property names
        :param skip_geometry: bool of whether to skip geometry (default False)
        :param q: full-text search term(s)
        :param filterq: CQL query as text string

        :returns: GeoJSON FeaturesCollection
        """
        LOGGER.debug(f"properties contains: {properties}")

        # NOTE: properties contains field keys plus extra params
        #       need to split them up here
        filtered_properties = []
        extra_params = {}
        for (key, value) in properties:
            if key in self.fields.keys():
                filtered_properties.append((key, value))
            else:
                extra_params[key] = value

        properties = filtered_properties

        # Check mandatory filter properties
        property_dict = dict(properties)
        if self.mandatory_properties:
            for mand_col in self.mandatory_properties:
                if mand_col == "bbox" and not bbox:
                    raise ProviderInvalidQueryError(
                        f"Missing mandatory filter property: {mand_col}"
                    )
                else:
                    if mand_col not in property_dict:
                        raise ProviderInvalidQueryError(
                            f"Missing mandatory filter property: {mand_col}"
                        )

        with DatabaseConnection(
            self.conn_dic,
            self.table,
            properties=self.properties,
            context="hits",
        ) as db:
            cursor = db.conn.cursor()

            where_dict = self._get_where_clauses(
                properties=properties,
                bbox=bbox,
                bbox_crs=self.storage_crs,
                sdo_param=self.sdo_param,
                sdo_operator=self.sdo_operator,
            )

            # Not dangerous to use self.table as substitution,
            # because of getFields ...
            sql_query = f"SELECT COUNT(1) AS hits \
                            FROM {self.table} \
                            {where_dict['clause']} #WHERE#"

            # Assign where_dict["properties"] to bind_variables
            bind_variables = {**where_dict["properties"]}

            # Default values for the process_query function (sql_manipulator)
            query_args = {
                "offset": offset,
                "limit": limit,
                "resulttype": resulttype,
                "bbox": bbox,
                "datetime_": datetime_,
                "properties": properties,
                "sortby": sortby,
                "skip_geometry": skip_geometry,
                "select_properties": select_properties,
                "crs_transform_spec": crs_transform_spec,
                "q": q,
                "language": language,
                "filterq": filterq,
            }

            # Apply the SQL manipulation plugin
            extra_params["geom"] = self.geom
            sql_query, bind_variables = self._process_query_with_sql_manipulator_sup(   # noqa: E501
                db, sql_query, bind_variables, extra_params, **query_args
            )

            try:
                cursor.execute(sql_query, bind_variables)
            except oracledb.Error as err:
                LOGGER.error(
                    f"Error executing sql_query: {sql_query}: {err}"
                )
                raise ProviderQueryError()

            hits = cursor.fetchone()[0]
            LOGGER.debug(f"hits: {str(hits)}")

        with DatabaseConnection(
            self.conn_dic, self.table, properties=self.properties
        ) as db:
            db.conn.outputtypehandler = self._output_type_handler

            cursor = db.conn.cursor()

            # Create column list.
            #   Uses columns field that was generated in the Connection class
            #   or the configured columns from the Yaml file.
            props = ", ".join(
                select_properties
                if select_properties
                else db.filtered_fields
            )

            where_dict = self._get_where_clauses(
                properties=properties,
                bbox=bbox,
                bbox_crs=self.storage_crs,
                sdo_param=self.sdo_param,
                sdo_operator=self.sdo_operator,
            )

            # Get correct SRID
            if crs_transform_spec is not None:
                source_crs = pyproj.CRS.from_wkt(
                    crs_transform_spec.source_crs_wkt
                )
                source_srid = get_srid(source_crs)

                target_crs = pyproj.CRS.from_wkt(
                    crs_transform_spec.target_crs_wkt
                )
                target_srid = get_srid(target_crs)
            else:
                source_srid = get_srid(self.storage_crs)
                target_srid = source_srid

            LOGGER.debug(f"source_srid: {source_srid}")
            LOGGER.debug(f"target_srid: {target_srid}")

            # Build geometry column call
            #   When a different output CRS is defined, the geometry
            #   geometry column would be transformed.
            if skip_geometry:
                geom = ""

            elif source_srid != target_srid:
                geom = f""", sdo_cs.transform(t1.{self.geom},
                                             :target_srid).get_geojson()
                             AS geometry """

                where_dict["properties"].update(
                    {"target_srid": int(target_srid)}
                )

            else:
                geom = f", t1.{self.geom}.get_geojson() AS geometry "

            orderby = self._get_orderby(sortby) if sortby else ""

            # Create paging and add placeholders for the
            # SQL manipulation class
            paging_bind = {}
            if limit > 0:
                sql_query = f"SELECT #HINTS# {props} \
                              {self._get_extra_columns_expression()} {geom} \
                              FROM {self.table} t1 #JOIN# \
                              {where_dict['clause']} #WHERE# \
                              {orderby} \
                              OFFSET :offset ROWS FETCH NEXT :limit ROWS ONLY"
                paging_bind = {"offset": offset, "limit": limit}
            else:
                sql_query = f"SELECT #HINTS# {props} \
                        {self._get_extra_columns_expression()} {geom} \
                        FROM {self.table} t1 #JOIN# \
                        {where_dict['clause']} #WHERE# \
                        {orderby}"

            # Create dictionary for sql bind variables
            bind_variables = {**where_dict["properties"], **paging_bind}

            # Apply the SQL manipulation plugin
            sql_query, bind_variables = self._process_query_with_sql_manipulator_sup(   # noqa: E501
                db, sql_query, bind_variables, extra_params, **query_args
            )

            LOGGER.debug(f"SQL Query: {sql_query}")
            LOGGER.debug(f"Bind variables: {bind_variables}")

            try:
                cursor.execute(sql_query, bind_variables)
            except oracledb.Error as err:
                LOGGER.error(f"Error executing sql_query: {sql_query}")
                LOGGER.error(err)
                raise ProviderQueryError()

            # Convert row resultset to dictionary
            columns = [col[0] for col in cursor.description]
            cursor.rowfactory = lambda *args: dict(zip(columns, args))

            row_data = cursor.fetchall()

            # Generate feature JSON
            features = [self._response_feature(rd) for rd in row_data]
            feature_collection = {
                "numberMatched": hits,
                "type": "FeatureCollection",
                "features": features,
            }

            return feature_collection

    def _get_previous(self, cursor, identifier):
        """
        Query previous ID given current ID

        :param identifier: feature id

        :returns: feature id
        """
        sql = f"SELECT {self.id_field} AS id \
                  FROM {self.table} \
                 WHERE ROWNUM = 1 \
                   AND {self.id_field} < :{self.id_field} \
                 ORDER BY {self.id_field} DESC"

        bind_variables = {self.id_field: identifier}

        LOGGER.debug(f"SQL Query: {sql}")
        LOGGER.debug(f"Bind variables: {str(bind_variables)}")

        cursor.execute(sql, bind_variables)

        item = cursor.fetchall()
        id = item[0][0] if item else None

        return id

    def _get_next(self, cursor, identifier):
        """
        Query next ID given current ID

        :param identifier: feature id

        :returns: feature id
        """
        sql = f"SELECT {self.id_field} AS id \
                  FROM {self.table} \
                 WHERE ROWNUM = 1 \
                   AND {self.id_field} > :{self.id_field} \
                 ORDER BY {self.id_field} ASC"

        bind_variables = {self.id_field: identifier}

        LOGGER.debug(f"SQL Query: {sql}")
        LOGGER.debug(f"Bind variables: {str(bind_variables)}")

        cursor.execute(sql, bind_variables)

        item = cursor.fetchall()
        id = item[0][0] if item else None

        return id

    def get(self, identifier, crs_transform_spec=None, **kwargs):
        """
        Query the provider for a specific
        feature id e.g: /collections/ocrl_lakes/items/1

        :param identifier: feature id

        :returns: GeoJSON FeaturesCollection
        """

        with DatabaseConnection(
            self.conn_dic, self.table, properties=self.properties
        ) as db:
            db.conn.outputtypehandler = self._output_type_handler

            cursor = db.conn.cursor()

            crs_dict = {}

            # Get correct SRIDs
            if crs_transform_spec is not None:
                source_crs = pyproj.CRS.from_wkt(
                    crs_transform_spec.source_crs_wkt
                )
                source_srid = get_srid(source_crs)

                target_crs = pyproj.CRS.from_wkt(
                    crs_transform_spec.target_crs_wkt
                )
                target_srid = get_srid(target_crs)

            else:
                source_srid = get_srid(self.storage_crs)
                target_srid = source_srid

            LOGGER.debug(f"source_srid: {source_srid}")
            LOGGER.debug(f"target_srid: {target_srid}")

            # Build geometry column call
            #   When a different output CRS is defined, the geometry
            #   geometry column would be transformed.
            if source_srid != target_srid:
                crs_dict = {"target_srid": target_srid}

                geom_sql = f""", sdo_cs.transform(t1.{self.geom},
                                             :target_srid).get_geojson()
                                    AS geometry """

            else:
                geom_sql = f", t1.{self.geom}.get_geojson() AS geometry "

            columns = ", ".join(db.filtered_fields)
            sql_query = f"SELECT {columns} \
                            {self._get_extra_columns_expression()} \
                            {geom_sql} \
                            FROM {self.table} t1 \
                           WHERE {self.id_field} = :in_id"

            bind_variables = {"in_id": identifier, **crs_dict}

            # SQL manipulation plugin
            if self.sql_manipulator:
                LOGGER.debug("sql_manipulator: " + self.sql_manipulator)
                manipulation_class = _class_factory(self.sql_manipulator)
                sql_query, bind_variables = manipulation_class.process_get(
                    db,
                    sql_query,
                    bind_variables,
                    self.sql_manipulator_options,
                    identifier,
                )

            LOGGER.debug(f"SQL Query: {sql_query}")
            LOGGER.debug(f"Identifier: {identifier}")

            try:
                cursor.execute(sql_query, bind_variables)
            except oracledb.Error as err:
                LOGGER.error(f"Error executing sql_query: {sql_query}")
                LOGGER.error(err)
                raise ProviderQueryError()

            # Convert row resultset to dictionary
            columns = [col[0] for col in cursor.description]
            cursor.rowfactory = lambda *args: dict(zip(columns, args))

            results = cursor.fetchall()

            row_data = None
            if results:
                row_data = results[0]
            feature = self._response_feature(row_data)

            if feature:
                previous_id = self._get_previous(cursor, identifier)
                if previous_id:
                    feature["prev"] = previous_id
                next_id = self._get_next(cursor, identifier)
                if next_id:
                    feature["next"] = self._get_next(cursor, identifier)
                return feature
            else:
                err = f"item identifier {identifier} not found"
                LOGGER.error(err)
                raise ProviderItemNotFoundError(err)

    def _response_feature(self, row_data):
        """
        Assembles GeoJSON output from DB query

        :param row_data: DB row result

        :returns: `dict` of GeoJSON Feature
        """

        if row_data:
            feature = {"type": "Feature"}

            if row_data.get("GEOMETRY"):
                feature["geometry"] = json.loads(row_data["GEOMETRY"])
            else:
                feature["geometry"] = None

            feature["properties"] = {
                key.lower(): value
                for (key, value) in row_data.items()
                if key != "GEOMETRY"
            }
            feature["id"] = feature["properties"][self.id_field]

            return feature
        else:
            return None

    def _response_feature_hits(self, hits):
        """Assembles GeoJSON/Feature number
        e.g: http://localhost:5000/collections/lakes/items?resulttype=hits

        :returns: GeoJSON FeaturesCollection
        """

        feature_collection = {"features": [], "type": "FeatureCollection"}
        feature_collection["numberMatched"] = hits

        return feature_collection

    def create(self, request_data):
        """
        Creates on record with the given data.

        :param request_data: Data of the record as Geojson
        :returns: ID of the created record
        """
        LOGGER.debug(f"Request data: {str(request_data)}")

        with DatabaseConnection(
            self.conn_dic, self.table, properties=self.properties
        ) as db:
            cursor = db.conn.cursor()

            columns = [*request_data.get("properties")]

            # Filter properties to get only columns who are
            # in the column list
            columns = [
                col
                for col in columns
                if col.lower() in db.filtered_fields
            ]

            # Filter function to get only properties who are
            # in the column list
            def filter_binds(pair):
                return pair[0].lower() in db.filtered_fields

            # Filter bind variables
            bind_variables = dict(
                filter(filter_binds, request_data.get("properties").items())
            )

            columns_str = ", ".join([col for col in columns])
            values_str = ", ".join([f":{col}" for col in columns])

            sql_query = f"""
                        INSERT INTO {self.table} (
                            {columns_str},
                            {self.geom}
                        )
                        VALUES (
                            {values_str},
                            sdo_util.from_geojson(:in_geometry, NULL, :srid)
                        )
                        RETURNING {self.id_field} INTO :out_id
                        """

            # Out bind variable for the id of the created row
            out_id = cursor.var(int)

            # Bind variable for the SDO_GEOMETRY type
            # in_geometry = self._get_sdo_from_geojson_geometry(
            #     db.conn, request_data.get("geometry").get("coordinates")[0]
            # )
            in_geometry = request_data.get("geometry")

            bind_variables = {
                **bind_variables,
                "out_id": out_id,
                "in_geometry": json.dumps(in_geometry),
                "srid": get_srid(self.storage_crs),
            }

            # SQL manipulation plugin
            if self.sql_manipulator:
                LOGGER.debug("sql_manipulator: " + self.sql_manipulator)
                manipulation_class = _class_factory(self.sql_manipulator)
                sql_query, bind_variables = manipulation_class.process_create(
                    db,
                    sql_query,
                    bind_variables,
                    self.sql_manipulator_options,
                    request_data,
                )

            LOGGER.debug(f"SQL Query: {sql_query}")
            LOGGER.debug(f"Bind variables: {bind_variables}")

            try:
                cursor.execute(sql_query, bind_variables)
                db.conn.commit()
            except oracledb.Error as err:
                LOGGER.error(f"Error executing sql_query: {sql_query}")
                LOGGER.error(err)

                db.conn.rollback()

                raise ProviderQueryError()

            identifier = out_id.getvalue()

        return identifier[0]

    def update(self, identifier, request_data):
        """
        Updates the record with the given identifier.

        :param identifier: ID of the record
        :param request_data: Data of the record as Geojson
        :returns: True
        """
        LOGGER.debug(f"Identifier: {identifier}")
        LOGGER.debug(f"Request data: {str(request_data)}")

        with DatabaseConnection(
            self.conn_dic, self.table, properties=self.properties
        ) as db:
            cursor = db.conn.cursor()

            columns = [*request_data.get("properties")]

            # Filter properties to get only columns who are
            # in the column list
            columns = [
                col
                for col in columns
                if col.lower() in db.filtered_fields
            ]

            # Filter function to get only properties who are
            # in the column list
            def filter_binds(pair):
                return pair[0].lower() in db.filtered_fields

            # Filter bind variables
            bind_variables = dict(
                filter(
                    filter_binds,
                    request_data.get("properties").items(),
                )
            )

            set_str = ", ".join([f" {col} = :{col}" for col in columns])

            sql_query = f"""
                        UPDATE {self.table}
                           SET {set_str}
                             , {self.geom} = sdo_util.from_geojson(
                                                 :in_geometry,
                                                 NULL,
                                                 :srid
                                             )
                         WHERE {self.id_field} = :in_id
                        """

            # Bind variable for the SDO_GEOMETRY type
            # in_geometry = self._get_sdo_from_geojson_geometry(
            #     db.conn, request_data.get("geometry").get("coordinates")[0]
            # )
            in_geometry = json.dumps(request_data.get("geometry"))

            bind_variables = {
                **bind_variables,
                "in_id": identifier,
                "in_geometry": in_geometry,
                "srid": get_srid(self.storage_crs),
            }

            # SQL manipulation plugin
            if self.sql_manipulator:
                LOGGER.debug("sql_manipulator: " + self.sql_manipulator)
                manipulation_class = _class_factory(self.sql_manipulator)
                sql_query, bind_variables = manipulation_class.process_update(
                    db,
                    sql_query,
                    bind_variables,
                    self.sql_manipulator_options,
                    identifier,
                    request_data,
                )

            LOGGER.debug(sql_query)
            LOGGER.debug(bind_variables)

            try:
                cursor.execute(sql_query, bind_variables)
                rowcount = cursor.rowcount
                db.conn.commit()
            except oracledb.Error as err:
                LOGGER.error(f"Error executing sql_query: {sql_query}")
                LOGGER.error(err)

                db.conn.rollback()

                raise ProviderQueryError()

        return rowcount == 1

    def delete(self, identifier):
        """
        Deletes the record with the given identifier.

        :param identifier: ID of the record
        :returns: True
        """

        LOGGER.debug(f"Identifier: {identifier}")

        with DatabaseConnection(
            self.conn_dic, self.table, properties=self.properties
        ) as db:
            cursor = db.conn.cursor()

            sql_query = f"DELETE FROM {self.table} \
                           WHERE {self.id_field} = :in_id"

            bind_variables = {
                "in_id": identifier,
            }

            # SQL manipulation plugin
            if self.sql_manipulator:
                LOGGER.debug("sql_manipulator: " + self.sql_manipulator)
                manipulation_class = _class_factory(self.sql_manipulator)
                sql_query, bind_variables = manipulation_class.process_delete(
                    db,
                    sql_query,
                    bind_variables,
                    self.sql_manipulator_options,
                    identifier,
                )

            LOGGER.debug(sql_query)
            LOGGER.debug(bind_variables)

            try:
                cursor.execute(sql_query, bind_variables)
                rowcount = cursor.rowcount
                db.conn.commit()
            except oracledb.Error as err:
                LOGGER.error(f"Error executing sql_query: {sql_query}")
                LOGGER.error(err)

                db.conn.rollback()

                raise ProviderQueryError()

        return rowcount == 1

    def _get_sdo_from_geojson_geometry(self, conn, geometry, srid=4326):
        """
        Get an filled Python object for Oracle Type SDO_GEOMETRY.

        :param conn: oracledb connection instance
        :param geometry: Ordinate Array from Geojson
        :param srid: SRID defaults to 4326 when not provided
        :return Python object instance:
        """
        gtype = 2003
        elemInfo = [1, 1003, 1]

        # Get Oracle types
        obj_type = conn.gettype("MDSYS.SDO_GEOMETRY")
        element_info_type_obj = conn.gettype("MDSYS.SDO_ELEM_INFO_ARRAY")
        ordinate_type_obj = conn.gettype("MDSYS.SDO_ORDINATE_ARRAY")

        obj = obj_type.newobject()
        obj.SDO_GTYPE = gtype
        obj.SDO_SRID = srid or 4326
        obj.SDO_ELEM_INFO = element_info_type_obj.newobject()
        obj.SDO_ELEM_INFO.extend(elemInfo)
        obj.SDO_ORDINATES = ordinate_type_obj.newobject()
        for coord in geometry:
            obj.SDO_ORDINATES.extend(coord)

        return obj


def _class_factory(
    module_class_string, super_cls: Optional[type] = None, **kwargs
):
    """
    Factory function for class instances.
    Used for dynamic loading of the SQL manipulation class.

    :param module_class_string: full name of the class to create an object of
    :param super_cls: expected super class for validity, None if bypass
    :param kwargs: parameters to pass
    :return instance of the given class:
    """
    module_name, class_name = module_class_string.rsplit(".", 1)
    module = importlib.import_module(module_name)
    LOGGER.debug(f"reading class {class_name} from module {module_name}")
    cls = getattr(module, class_name)
    if super_cls is not None:
        assert issubclass(
            cls, super_cls
        ), f"class {class_name} should inherit from {super_cls.__name__}"
    LOGGER.debug(f"initialising {class_name} with params {kwargs}")
    obj = cls(**kwargs)
    return obj



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2026 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging

from kafka import errors, KafkaProducer

from pygeoapi.pubsub.base import BasePubSubClient, PubSubClientConnectionError
from pygeoapi.util import to_json

LOGGER = logging.getLogger(__name__)


class KafkaPubSubClient(BasePubSubClient):
    """Kafka client"""

    def __init__(self, publisher_def):
        """
        Initialize object

        :param publisher_def: provider definition

        :returns: pygeoapi.pubsub.kafka.KafkaPubSubClient
        """

        super().__init__(publisher_def)
        self.name = 'Kafka'
        self.type = 'kafka'
        self.sasl_mechanism = publisher_def.get('sasl.mechanism', 'PLAIN')
        self.security_protocol = publisher_def.get('security.protocol', 'SASL_SSL')  # noqa

        msg = f'Initializing to broker {self.broker_safe_url} with id {self.client_id}'  # noqa
        LOGGER.debug(msg)

    def connect(self) -> None:
        """
        Connect to an Kafka broker

        :returns: None
        """

        args = {
            'bootstrap_servers': f'{self.broker_url.hostname}:{self.broker_url.port}',  # noqa
            'client_id': self.client_id,
            'value_serializer': lambda v: to_json(v).encode('utf-8')
        }
        if None not in [self.broker_url.username, self.broker_url.password]:
            args.update({
                'security.protocol': self.security_protocol,
                'sasl.mechanism': self.sasl_mechanism,
                'sasl.username': self.broker_url.username,
                'sasl.password': self.broker_url.password
            })

        LOGGER.debug('Creating Kafka producer')
        try:
            self.producer = KafkaProducer(**args)
        except errors.NoBrokersAvailable as err:
            raise PubSubClientConnectionError(err)

    def pub(self, channel: str, message: str) -> bool:
        """
        Publish a message to a broker/channel

        :param channel: `str` of topic
        :param message: `str` of message

        :returns: `bool` of publish result
        """

        LOGGER.debug(f'Publishing to broker {self.broker_safe_url}')
        LOGGER.debug(f'Channel: {channel}')
        LOGGER.debug(f'Message: {message}')
        LOGGER.debug('Sanitizing channel for HTTP')
        channel = channel.replace('/', '-')
        channel = channel.replace(':', '-')
        LOGGER.debug(f'Sanitized channel for Kafka: {channel}')

        self.producer.send(channel, value=message)
        self.producer.flush()

    def __repr__(self):
        return f'<HTTPPubSubClient> {self.broker_safe_url}'



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2025 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging
from urllib.parse import urlencode

from owslib import fes
from owslib.csw import CatalogueServiceWeb
from owslib.ows import ExceptionReport

from pygeoapi.crs import crs_transform
from pygeoapi.provider.base import (BaseProvider, ProviderConnectionError,
                                    ProviderInvalidQueryError,
                                    ProviderItemNotFoundError,
                                    ProviderQueryError)
from pygeoapi.util import bbox2geojsongeometry, get_typed_value

LOGGER = logging.getLogger(__name__)


class CSWFacadeProvider(BaseProvider):
    """CSW Facade provider"""

    def __init__(self, provider_def):
        """
        Initialize object

        :param provider_def: provider definition

        :returns: pygeoapi.provider.csw_facade.CSWFacadeProvider
        """

        super().__init__(provider_def)

        self.record_mappings = {
            'type': ('dc:type', 'type'),
            'title': ('dc:title', 'title'),
            'description': ('dct:abstract', 'abstract'),
            'keywords': ('dc:subject', 'subjects'),
            'date': ('dc:date', 'date'),
            'created': ('dct:created', 'created'),
            'updated': ('dct:modified', 'modified'),
            'rights': ('dc:rights', 'rights'),
            'language': ('dc:language', 'language')
        }

        self._fields = {}
        self.get_fields()

    def get_fields(self):
        """
         Get provider field information (names, types)

        :returns: dict of fields
        """

        if not self._fields:
            date_fields = ['date', 'created', 'updated']

            for key in self.record_mappings.keys():
                LOGGER.debug(f'key: {key}')
                self._fields[key] = {'type': 'string'}

                if key in date_fields:
                    self._fields[key]['format'] = 'date-time'

        return self._fields

    def get_domains(self, properties=[], current=False) -> tuple:
        """
        Get domains from dataset

        :param properties: `list` of property names
        :param current: `bool` of whether to provide list of live
                        values (default `False`)

        :returns: `tuple` of domains and whether they are based on the
                  current/live dataset
        """

        LOGGER.debug(f'Querying CSW: {self.data}')
        records = self.query()
        domains = {}

        if properties:
            keys = properties
        else:
            keys = records['features'][0]['properties'].keys()

        csw = self._get_csw()

        for key in keys:
            csw.getdomain(key, dtype='property')
            domains[key] = csw.results['values']

        return domains, True

    @crs_transform
    def query(self, offset=0, limit=10, resulttype='results',
              bbox=[], datetime_=None, properties=[], sortby=[],
              select_properties=[], skip_geometry=False, q=None, **kwargs):
        """
        CSW GetRecords query

        :param offset: starting record to return (default 0)
        :param limit: number of records to return (default 10)
        :param resulttype: return results or hit limit (default results)
        :param bbox: bounding box [minx,miny,maxx,maxy]
        :param datetime_: temporal (datestamp or extent)
        :param properties: list of tuples (name, value)
        :param sortby: list of dicts (property, order)
        :param select_properties: list of property names
        :param skip_geometry: bool of whether to skip geometry (default False)
        :param q: full-text search term(s)

        :returns: `dict` of GeoJSON FeatureCollection
        """

        constraints = []

        response = {
            'type': 'FeatureCollection',
            'features': []
        }

        LOGGER.debug('Processing query parameters')

        if bbox:
            LOGGER.debug('Processing bbox parameter')
            LOGGER.debug('Swapping coordinate axis order from xy to yx')
            bbox2 = [bbox[1], bbox[0], bbox[3], bbox[2]]
            constraints.append(fes.BBox(bbox2))

        if datetime_:
            date_property = self.record_mappings[self.time_field][0]
            LOGGER.debug('Processing datetime parameter')
            if '/' in datetime_:
                begin, end = datetime_.split('/')
                LOGGER.debug('Processing time extent')
                constraints.append(fes.PropertyIsGreaterThan(date_property, begin))  # noqa
                constraints.append(fes.PropertyIsLessThan(date_property, end))
            else:
                LOGGER.debug('Processing time instant')
                constraints.append(fes.PropertyIsEqualTo(date_property,
                                                         datetime_))

        for p in properties:
            LOGGER.debug(f'Processing property {p} parameter')
            if p[0] not in list(self.record_mappings.keys()):
                msg = f'Invalid property: {p[0]}'
                LOGGER.error(msg)
                raise ProviderInvalidQueryError(user_msg=msg)

            prop = self.record_mappings[p[0]][0]
            constraints.append(fes.PropertyIsEqualTo(prop, p[1]))

        if q is not None:
            LOGGER.debug('Processing q parameter')
            anytext = fes.PropertyIsLike(propertyname='csw:AnyText', literal=q,
                                         escapeChar='\\', singleChar='?',
                                         wildCard='*')
            constraints.append(anytext)

        if sortby:
            LOGGER.debug('Processing sortby parameter')
            sorts = []
            sort_orders = {
                '+': 'ASC',
                '-': 'DESC'
            }
            for s in sortby:
                sorts.append(fes.SortProperty(
                    self.record_mappings[s['property']][0],
                    sort_orders[s['order']]))
            sortby2 = fes.SortBy(sorts)
        else:
            sortby2 = None

        if len(constraints) > 1:
            constraints = [fes.And(constraints)]

        LOGGER.debug(f'Querying CSW: {self.data}')
        csw = self._get_csw()
        try:
            csw.getrecords2(esn='full', maxrecords=limit, startposition=offset,
                            constraints=constraints, sortby=sortby2,
                            resulttype=resulttype)
        except ExceptionReport as err:
            msg = f'CSW error {err}'
            LOGGER.error(msg)
            raise ProviderQueryError(msg)

        response['numberMatched'] = csw.results['matches']
        response['numberReturned'] = csw.results['returned']
        LOGGER.debug(f"Found {response['numberMatched']} records")
        LOGGER.debug(f"Returned {response['numberReturned']} records")

        LOGGER.debug('Building result set')
        for record in csw.records.values():
            response['features'].append(self._owslibrecord2record(record))

        return response

    @crs_transform
    def get(self, identifier, **kwargs):
        """
        CSW GetRecordById query

        :param identifier: feature id

        :returns: dict of single GeoJSON feature
        """

        csw = self._get_csw()
        csw.getrecordbyid([identifier], esn='full')

        if not csw.records:
            err = f'item {identifier} not found'
            LOGGER.error(err)
            raise ProviderItemNotFoundError(err)

        record_key = list(csw.records.keys())[0]

        return self._owslibrecord2record(csw.records[record_key])

    def _get_csw(self) -> CatalogueServiceWeb:
        """
        Helper function to lazy load a CSW

        returns: `owslib.csw.CatalogueServiceWeb`
        """

        try:
            return CatalogueServiceWeb(self.data)
        except Exception as err:
            err = f'CSW connection error: {err}'
            LOGGER.error(err)
            raise ProviderConnectionError(err)

    def _gen_getrecordbyid_link(self, identifier: str,
                                csw_version: str = '2.0.2') -> dict:
        """
        Helper function to generate a CSW GetRecordById URL

        :param identifier: `str` of record identifier
        :param csw_version: `str` of CSW version (default is `2.0.2`)

        :returns: `dict` of link object of GetRecordById URL
        """

        params = {
            'service': 'CSW',
            'version': csw_version,
            'request': 'GetRecordById',
            'id': identifier
        }

        return {
            'rel': 'alternate',
            'type': 'application/xml',
            'title': 'This document as XML',
            'href': f'{self.data}?{urlencode(params)}',
        }

    def _owslibrecord2record(self, record):
        LOGGER.debug(f'Transforming {record.identifier}')
        feature = {
            'id': record.identifier,
            'type': 'Feature',
            'geometry': None,
            'time': record.date or None,
            'properties': {},
            'links': [
                self._gen_getrecordbyid_link(record.identifier)
            ]
        }

        LOGGER.debug('Processing record mappings to properties')
        for key, value in self.record_mappings.items():
            prop_value = getattr(record, value[1])
            if prop_value not in [None, [], '']:
                feature['properties'][key] = prop_value

        if record.bbox is not None:
            LOGGER.debug('Adding bbox')
            bbox = [
                get_typed_value(record.bbox.minx),
                get_typed_value(record.bbox.miny),
                get_typed_value(record.bbox.maxx),
                get_typed_value(record.bbox.maxy)
            ]
            feature['geometry'] = bbox2geojsongeometry(bbox)

        if record.references:
            LOGGER.debug('Adding references as links')
            for link in record.references:
                feature['links'].append({
                    'title': link['scheme'],
                    'href': link['url']
                })
        if record.uris:
            LOGGER.debug('Adding URIs as links')
            for link in record.uris:
                feature['links'].append({
                    'title': link['name'],
                    'href': link['url']
                })

        return feature

    def __repr__(self):
        return f'<CSWFacadeProvider> {self.data}'



# =================================================================
#
# Authors: Antonio Cerciello <anto.nio.cerciello@gmail.com>
#          Francesco Bartoli <xbartolone@gmail.com>
#          Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2022 Antonio Cerciello
# Copyright (c) 2025 Francesco Bartoli
# Copyright (c) 2025 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from datetime import datetime
from enum import Enum
import json
from pathlib import Path
from typing import List, Optional

import pydantic
from pydantic import BaseModel

from pygeoapi.util import DEFINITIONSDIR

TMS_DIR = DEFINITIONSDIR / 'tiles'


class TilesMetadataFormat(str, Enum):
    # Tile Set Metadata
    JSON = "JSON"
    JSONLD = "JSONLD"
    # TileJSON 3.0
    TILEJSON = "TILEJSON"
    # HTML (default)
    HTML = "HTML"


# Tile Set Metadata Enums
class AccessConstraintsEnum(str, Enum):
    UNCLASSIFIED = "unclassified"
    RESTRICTED = "restricted"
    CONFIDENTIAL = "confidential"
    SECRET = "secret"
    TOPSECRET = "topSecret"


class DataTypeEnum(str, Enum):
    MAP = "map"
    VECTOR = "vector"
    COVERAGE = "coverage"


class GeometryDimensionEnum(int, Enum):
    POINTS = 0
    CURVES = 1
    SURFACES = 2
    SOLIDS = 3


class TileMatrixSetEnumType(BaseModel):
    tileMatrixSet: str
    tileMatrixSetURI: str
    crs: str
    title: str
    orderedAxes: List[str]
    wellKnownScaleSet: str
    tileMatrices: List[dict]


class TileMatrixSetLoader:
    def __init__(self, directory: Path):
        self.directory = directory

    def load_from_file(self, filename: str) -> TileMatrixSetEnumType:
        """
        Load a single TMS JSON file.

        :param filename: filename of TMS

        :returns: `TileMatrixSetEnumType` of TMS
        """

        filepath = self.directory / filename

        with filepath.open(encoding='utf-8') as fh:
            data = json.load(fh)

        return TileMatrixSetEnumType(
            tileMatrixSet=data["id"],
            tileMatrixSetURI=data["uri"],
            crs=data["crs"],
            title=data["title"],
            orderedAxes=data["orderedAxes"],
            wellKnownScaleSet=data.get("wellKnownScaleSet", ""),
            tileMatrices=data["tileMatrices"]
        )

    def create_enum(self) -> Enum:
        """
        Create an Enum with all TileMatrixSets in the directory.

        :returns: `Enum` of `TileMatrixSetEnum`
        """

        members = {}
        for filepath in self.directory.glob("*.json"):
            tms = self.load_from_file(filepath.name)
            enum_name = tms.tileMatrixSet.upper().replace("-", "").replace(" ", "")  # noqa
            members[enum_name] = tms

        return Enum("TileMatrixSetEnum", members)


tms_loader = TileMatrixSetLoader(TMS_DIR)
TileMatrixSetEnum = tms_loader.create_enum()


# Tile Set Metadata Sub Types
class TileMatrixLimitsType(BaseModel):
    tileMatrix: str
    minTileRow: int
    maxTileRow: int
    minTileCol: int
    maxTileCol: int


class TwoDBoundingBoxType(BaseModel):
    lowerLeft: List[float]
    upperRight: List[float]
    crs: Optional[str] = None


class LinkType(BaseModel):
    href: str
    rel: Optional[str] = None
    type_: Optional[str] = None
    hreflang: Optional[str] = None
    title: Optional[str] = None
    length: Optional[int] = None


class GeospatialDataType(BaseModel):
    id: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None
    keywords: Optional[List[str]] = None
    dataType: DataTypeEnum = DataTypeEnum.VECTOR
    geometryDimension: Optional[GeometryDimensionEnum] = None
    maxTileMatrix: Optional[str] = None
    minTileMatrix: Optional[str] = None
    minScaleDenominator: Optional[float] = None
    maxScaleDenominator: Optional[float] = None
    minCellSize: Optional[float] = None
    maxCellSize: Optional[float] = None
    boundingBox: Optional[TwoDBoundingBoxType] = None
    links: Optional[LinkType] = None
    propertiesSchema: Optional[dict] = None


class StyleType(BaseModel):
    id: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None
    keywords: Optional[List[str]] = None
    links: Optional[LinkType] = None


class TilePointType(BaseModel):
    crs: str
    coordinates: Optional[List[float]] = None
    scaleDenominator: Optional[float] = None
    cellSize: Optional[float] = None
    # CodeType as adaptation of MD_Identifier class ISO 19115
    tileMatrix: str
    cellSize: Optional[str] = None


class TileSetMetadata(BaseModel):
    # A title for this tileset
    title: Optional[str] = None
    # Brief narrative description of this tile set
    description: Optional[str] = None
    # keywords about this tileset
    keywords: Optional[List[str]] = None
    # Version of the Tile Set. Changes if the data behind the tiles
    # has been changed
    version: Optional[str] = None
    # Useful information to contact the authors or custodians for the Tile Set
    pointOfContact: Optional[str] = None
    # Short reference to recognize the author or provider
    attribution: Optional[str] = None
    # License applicable to the tiles
    license_: Optional[str] = None
    # Restrictions on the availability of the Tile Set that the user needs to
    # be aware of before using or redistributing the Tile Set
    accessConstraints: Optional[AccessConstraintsEnum] = AccessConstraintsEnum.UNCLASSIFIED  # noqa
    # Media types available for the tiles
    mediaTypes:  Optional[List[str]] = None
    # Type of data represented in the tileset
    dataType: DataTypeEnum = DataTypeEnum.VECTOR
    # Limits for the TileRow and TileCol values for each TileMatrix in the
    # tileMatrixSet. If missing, there are no limits other that the ones
    # imposed by the TileMatrixSet. If present the TileMatrices listed are
    # limited and the rest not available at all
    tileMatrixSetLimits: Optional[TileMatrixLimitsType] = None
    # Coordinate Reference System (CRS)
    crs: Optional[str] = None
    # Epoch of the Coordinate Reference System (CRS)
    epoch: Optional[int] = None
    # Minimum bounding rectangle surrounding the tile matrix set, in the
    # supported CRS
    boundingBox: Optional[TwoDBoundingBoxType] = None
    # When the Tile Set was first produced
    created: Optional[datetime] = None
    # Last Tile Set change/revision
    updated: Optional[datetime] = None
    layers: Optional[GeospatialDataType] = None
    # Style involving all layers used to generate the tileset
    style: Optional[StyleType] = None
    # Location of a tile that nicely represents the tileset.
    # Implementations may use this center value to set the default location
    # or to present a representative tile in a user interface
    centerPoint: Optional[TilePointType] = None
    # Tile matrix set definition
    tileMatrixSet: Optional[str] = None
    # Reference to a Tile Matrix Set on an official source
    tileMatrixSetURI: Optional[str] = None
    # Links to related resources.
    links: Optional[List[LinkType]] = None


if pydantic.VERSION.startswith('1'):
    def _dump(self, *, exclude_none: bool = False, **kwargs):
        return self.dict(exclude_none=exclude_none, **kwargs)

    TileSetMetadata.model_dump = _dump



# =================================================================

# Authors: Tom Kralidis <tomkralidis@gmail.com>
#          Francesco Bartoli <xbartolone@gmail.com>
#          Sander Schaminee <sander.schaminee@geocat.net>
#          John A Stevenson <jostev@bgs.ac.uk>
#          Colin Blackburn <colb@bgs.ac.uk>
#          Ricardo Garcia Silva <ricardo.garcia.silva@geobeyond.it>
#          Bernhard Mallinger <bernhard.mallinger@eox.at>
#
# Copyright (c) 2026 Tom Kralidis
# Copyright (c) 2025 Francesco Bartoli
# Copyright (c) 2022 John A Stevenson and Colin Blackburn
# Copyright (c) 2023 Ricardo Garcia Silva
# Copyright (c) 2024 Bernhard Mallinger
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from copy import deepcopy
from http import HTTPStatus
import json
import logging
from typing import Any, Tuple, Union
from urllib.parse import urlencode

from shapely import from_geojson

from pygeoapi import l10n
from pygeoapi.formats import FORMAT_TYPES, F_JSON, F_HTML
from pygeoapi import api as ogc_api
from pygeoapi.api import itemtypes as itemtypes_api
from pygeoapi.plugin import load_plugin

from pygeoapi.provider import get_provider_by_type
from pygeoapi.provider.base import (
    ProviderConnectionError, ProviderNotFoundError, ProviderTypeError
)
from pygeoapi.util import (filter_dict_by_key_value, get_current_datetime,
                           render_j2_template, to_json)

from . import APIRequest, API


LOGGER = logging.getLogger(__name__)


CONFORMANCE_CLASSES = []


# TODO: no tests for this?
def get_stac_root(api: API, request: APIRequest) -> Tuple[dict, int, str]:
    """
    Provide STAC root page

    :param request: APIRequest instance with query params

    :returns: tuple of headers, status code, content
    """
    headers = request.get_response_headers(**api.api_headers)

    id_ = 'pygeoapi-stac'
    stac_version = '1.0.0-rc.2'
    stac_url = f'{api.base_url}/stac'

    content = {
        'id': id_,
        'type': 'Catalog',
        'stac_version': stac_version,
        'title': l10n.translate(
            api.config['metadata']['identification']['title'],
            request.locale),
        'description': l10n.translate(
            api.config['metadata']['identification']['description'],
            request.locale),
        'links': []
    }

    stac_collections = filter_dict_by_key_value(api.config['resources'],
                                                'type', 'stac-collection')

    for key, value in stac_collections.items():
        try:
            _ = load_plugin('provider', get_provider_by_type(
                            value['providers'], 'stac'))
        except ProviderTypeError:
            LOGGER.debug('Not a STAC-based provider; skipping')
            continue

        content['links'].append({
            'rel': 'child',
            'href': f'{stac_url}/{key}?f={F_JSON}',
            'type': FORMAT_TYPES[F_JSON],
            'title': key,
            'description': value['description']
        })
        content['links'].append({
            'rel': 'child',
            'href': f'{stac_url}/{key}',
            'type': FORMAT_TYPES[F_HTML],
            'title': key,
            'description': value['description']
        })

    if request.format == F_HTML:  # render
        content = render_j2_template(
            api.tpl_config, api.config['server']['templates'],
            'stac/collection.html', content, request.locale)

        return headers, HTTPStatus.OK, content

    return headers, HTTPStatus.OK, to_json(content, api.pretty_print)


# TODO: no tests for this?
def get_stac_path(api: API, request: APIRequest,
                  path: str) -> Tuple[dict, int, str]:
    """
    Provide STAC resource path

    :param request: APIRequest instance with query params

    :returns: tuple of headers, status code, content
    """
    headers = request.get_response_headers(**api.api_headers)

    dataset = None
    LOGGER.debug(f'Path: {path}')
    dir_tokens = path.split('/')
    if dir_tokens:
        dataset = dir_tokens[0]

    stac_collections = filter_dict_by_key_value(api.config['resources'],
                                                'type', 'stac-collection')

    if dataset not in stac_collections:
        msg = 'Collection not found'
        return api.get_exception(HTTPStatus.NOT_FOUND, headers,
                                 request.format, 'NotFound', msg)

    LOGGER.debug('Loading provider')
    try:
        p = load_plugin('provider', get_provider_by_type(
            stac_collections[dataset]['providers'], 'stac'))
    except ProviderConnectionError:
        msg = 'connection error (check logs)'
        return api.get_exception(
            HTTPStatus.INTERNAL_SERVER_ERROR, headers,
            request.format, 'NoApplicableCode', msg)

    id_ = f'{dataset}-stac'
    stac_version = '1.0.0-rc.2'

    content = {
        'id': id_,
        'type': 'Catalog',
        'stac_version': stac_version,
        'description': l10n.translate(
            stac_collections[dataset]['description'], request.locale),
        'links': []
    }
    try:
        stac_data = p.get_data_path(
            f'{api.base_url}/stac',
            path,
            path.replace(dataset, '', 1)
        )
    except ProviderNotFoundError:
        msg = 'resource not found'
        return api.get_exception(HTTPStatus.NOT_FOUND, headers,
                                 request.format, 'NotFound', msg)
    except Exception:
        msg = 'data query error'
        return api.get_exception(
            HTTPStatus.INTERNAL_SERVER_ERROR, headers,
            request.format, 'NoApplicableCode', msg)

    if isinstance(stac_data, dict):
        content.update(stac_data)
        content['links'].extend(
            stac_collections[dataset].get('links', []))

        if request.format == F_HTML:  # render
            content['path'] = path
            if 'assets' in content:  # item view
                if content['type'] == 'Collection':
                    content = render_j2_template(
                        api.tpl_config, api.config['server']['templates'],
                        'stac/collection_base.html',
                        content, request.locale)
                elif content['type'] == 'Feature':
                    content = render_j2_template(
                        api.tpl_config, api.config['server']['templates'],
                        'stac/item.html', content, request.locale)
                else:
                    msg = f'Unknown STAC type {content.type}'
                    return api.get_exception(
                        HTTPStatus.INTERNAL_SERVER_ERROR,
                        headers,
                        request.format,
                        'NoApplicableCode',
                        msg)
            else:
                content = render_j2_template(
                        api.tpl_config, api.config['server']['templates'],
                        'stac/catalog.html', content, request.locale)

            return headers, HTTPStatus.OK, content

        return headers, HTTPStatus.OK, to_json(content, api.pretty_print)

    else:  # send back file
        headers.pop('Content-Type', None)
        return headers, HTTPStatus.OK, stac_data


def landing_page(api: API,
                 request: APIRequest) -> Tuple[dict, int, str]:
    """
    Provide API landing page

    :param request: A request object

    :returns: tuple of headers, status code, content
    """

    request._format = F_JSON

    headers, status, content = ogc_api.landing_page(api, request)

    content = json.loads(content)

    content['id'] = 'pygeoapi-catalogue'
    content['stac_version'] = '1.0.0'
    content['conformsTo'] = [
        'https://api.stacspec.org/v1.0.0/core',
        'https://api.stacspec.org/v1.0.0/item-search',
        'https://api.stacspec.org/v1.0.0/item-search#sort'
    ]
    content['type'] = 'Catalog'

    content['links'] = [{
        'rel': request.get_linkrel(F_JSON),
        'type': FORMAT_TYPES[F_JSON],
        'title': l10n.translate('This document as JSON', request.locale),
        'href': f"{api.base_url}/stac-api?f={F_JSON}"
    }, {
        'rel': 'root',
        'type': FORMAT_TYPES[F_JSON],
        'title': l10n.translate('This document as JSON', request.locale),
        'href': f"{api.base_url}/stac-api?f={F_JSON}"
    }, {
        'rel': 'service-desc',
        'type': 'application/vnd.oai.openapi+json;version=3.0',
        'title': l10n.translate('The OpenAPI definition as JSON', request.locale),  # noqa
        'href': f"{api.base_url}/openapi"
    }, {
        'rel': 'service-doc',
        'type': FORMAT_TYPES[F_HTML],
        'title': l10n.translate('The OpenAPI definition as HTML', request.locale),  # noqa
        'href': f"{api.base_url}/openapi?f={F_HTML}",
        'hreflang': api.default_locale
    }, {
        'rel': 'search',
        'type': FORMAT_TYPES[F_JSON],
        'title': l10n.translate('STAC API search', request.locale),
        'href': f"{api.base_url}/stac-api//search?f={F_JSON}"
    }]

    return headers, status, to_json(content, api.pretty_print)


def search(api: API, request: Union[APIRequest, Any]) -> Tuple[dict, int, str]:
    """
    STAC API Queries stac-collection

    :param request: A request object
    :param dataset: dataset name

    :returns: tuple of headers, status code, content
    """

    stac_api_collections = {}

    request._format = F_JSON

    headers = request.get_response_headers(**api.api_headers)

    LOGGER.debug('Checking for STAC collections')
    collections = filter_dict_by_key_value(api.config['resources'],
                                           'type', 'stac-collection')

    if not collections:
        return api.get_exception(
            HTTPStatus.NOT_IMPLEMENTED, headers, F_JSON, 'NotImplemented',
            'No configured STAC searchable collection')

    LOGGER.debug('Checking for STAC collections with features or records')
    for key, value in collections.items():
        found_collection = False
        for fr in ['feature', 'record']:
            try:
                _ = get_provider_by_type(value['providers'], fr)
                found_collection = True
                break
            except ProviderTypeError:
                pass

        if found_collection:
            stac_api_collections[key] = value

    if not stac_api_collections:
        msg = 'No STAC API collections configured'
        return api.get_exception(HTTPStatus.INTERNAL_SERVER_ERROR, headers,
                                 request.format, 'NotApplicable', msg)

    if request.data:
        LOGGER.debug('Intercepting STAC POST request into query args')
        request_data = json.loads(request.data)
        request_params = deepcopy(dict(request.params))

        for qp in ['bbox', 'datetime', 'limit', 'offset']:
            if qp in request_data:
                if qp == 'bbox' and isinstance(request_data[qp], list):
                    request_params[qp] = ','.join(str(b) for b in request_data[qp])  # noqa
                else:
                    request_params[qp] = request_data[qp]

        request._args = request_params
        request._data = None

    stac_api_response = {
        'type': 'FeatureCollection',
        'features': [],
        'numberMatched': 0,
        'links': []
    }

    for key, value in stac_api_collections.items():
        api.config['resources'][key]['type'] = 'collection'
        headers, status, content = itemtypes_api.get_collection_items(
            api, request, key)
        api.config['resources'][key]['type'] = 'stac-collection'

        if status != HTTPStatus.OK:
            return headers, status, to_json(content, api.pretty_print)

        content = json.loads(content)
        stac_api_response['numberMatched'] += content.get('numberMatched', 0)

        if len(content.get('features', [])) > 0:
            for feature in content['features']:
                if 'stac_version' not in feature:
                    feature['stac_version'] = '1.0.0'
                feature['properties'].update(get_temporal(feature))
                stac_api_response['features'].append(feature)

                if feature.get('geometry') is not None and 'bbox' not in feature:  # noqa
                    geom = from_geojson(json.dumps(feature['geometry']))
                    feature['bbox'] = geom.bounds

                for la in ['links', 'assets']:
                    if feature.get(la) is None:
                        feature[la] = []

    stac_api_response['numberReturned'] = len(stac_api_response['features'])

    stac_api_response['links'].append({
        'rel': 'root',
        'type': FORMAT_TYPES[F_JSON],
        'title': l10n.translate('STAC API landing page', request.locale),
        'href': f"{api.base_url}/stac-api?f={F_JSON}"
    })

    LOGGER.debug('Generating paging links')

    next_link = False
    prev_link = False
    request_params = deepcopy(dict(request._args))
    limit = itemtypes_api.evaluate_limit(
        request_params.get('limit'),
        api.config['server'].get('limits', {}), {})
    offset = int(request_params.get('offset', 0))

    if stac_api_response.get('numberMatched', -1) > (limit + offset):
        next_link = True
    elif len(stac_api_response['features']) == limit:
        next_link = True

    if offset > 0:
        prev_link = True

    if prev_link:
        request_params['offset'] = max(0, offset - limit)
        if request_params['offset'] == 0:
            request_params.pop('offset')

        request_params_qs = urlencode(request_params)

        stac_api_response['links'].append({
            'rel': 'prev',
            'type': FORMAT_TYPES[F_JSON],
            'title': l10n.translate('Items (prev)', request.locale),
            'href': f"{api.base_url}/stac-api/search?{request_params_qs}"
        })

    if next_link:
        request_params['offset'] = offset + limit
        request_params_qs = urlencode(request_params)

        stac_api_response['links'].append({
            'rel': 'next',
            'type': FORMAT_TYPES[F_JSON],
            'title': l10n.translate('Items (next)', request.locale),
            'href': f"{api.base_url}/stac-api/search?{request_params_qs}"
        })

    return headers, HTTPStatus.OK, to_json(stac_api_response, api.pretty_print)


def get_oas_30(cfg: dict, locale: str) -> tuple[list[dict[str, str]], dict[str, dict]]:  # noqa
    """
    Get OpenAPI fragments

    :param cfg: `dict` of configuration
    :param locale: `str` of locale

    :returns: `tuple` of `list` of tag objects, and `dict` of path objects
    """

    LOGGER.debug('setting up STAC')
    stac_collections = filter_dict_by_key_value(cfg['resources'],
                                                'type', 'stac-collection')
    paths = {}
    if stac_collections:
        paths['/stac/catalog'] = {
            'get': {
                'summary': 'SpatioTemporal Asset Catalog',
                'description': 'SpatioTemporal Asset Catalog',
                'tags': ['stac'],
                'operationId': 'getStacCatalog',
                'parameters': [],
                'responses': {
                    '200': {'$ref': '#/components/responses/200'},
                    'default': {'$ref': '#/components/responses/default'}
                }
            }
        }
    return [{'name': 'stac'}], {'paths': paths}


def get_temporal(feature: dict) -> dict:
    """
    Helper function to try and derive a useful temporal
    definition on a non-STAC item

    :param feature: `dict` of GeoJSON feature

    :returns: `dict` of `datetime` or `start_datetime` and `end_datetime`
    """

    value = {}

    datetime_ = feature['properties'].get('datetime')
    start_datetime = feature['properties'].get('start_datetime')
    end_datetime = feature['properties'].get('end_datetime')

    if datetime_ is None and None not in [start_datetime, end_datetime]:
        LOGGER.debug('Temporal range partially exists')
    elif datetime_ is not None:
        LOGGER.debug('Temporal instant exists')

    LOGGER.debug('Attempting to derive temporal from GeoJSON feature')
    LOGGER.debug(feature)
    if feature.get('time') is not None:
        if feature['time'].get('timestamp') is not None:
            value['datetime'] = feature['time']['timestamp']
        if feature['time'].get('interval') is not None:
            value['start_datetime'] = feature['time']['interval'][0]
            value['end_datetime'] = feature['time']['interval'][1]

    if feature['properties'].get('created') is not None:
        value['datetime'] = feature['properties']['created']

    if not value:
        value['datetime'] = get_current_datetime()

    return value



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2026 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging

LOGGER = logging.getLogger(__name__)


class BaseFormatter:
    """generic Formatter ABC"""

    def __init__(self, formatter_def: dict):
        """
        Initialize object

        param formatter_def: formatter definition

        :returns: pygeoapi.formatter.base.BaseFormatter
        """

        self.extension = None
        self.f = None
        self.mimetype = None

        try:
            self.name = formatter_def['name']
        except KeyError:
            raise RuntimeError('name is required')

        self.geom = formatter_def.get('geom', False)
        self.attachment = formatter_def.get('attachment', False)

    def write(self, options: dict = {}, data: dict | None = None) -> str:
        """
        Generate data in specified format

        :param options: formatting options
        :param data: dict representation of GeoJSON object

        :returns: string representation of format
        """

        raise NotImplementedError()

    def __repr__(self):
        return f'<BaseFormatter> {self.name}'


class FormatterGenericError(Exception):
    """formatter generic error"""
    pass


class FormatterSerializationError(FormatterGenericError):
    """formatter serialization error"""
    pass



# =================================================================
#
# Authors: Francesco Bartoli <francesco.bartoli@geobeyond.it>
#          Luca Delucchi <lucadeluge@gmail.com>
#          Krishna Lodha <krishnaglodha@gmail.com>
#          Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2025 Francesco Bartoli
# Copyright (c) 2022 Luca Delucchi
# Copyright (c) 2022 Krishna Lodha
# Copyright (c) 2022 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import os
from pathlib import Path
import sys

from pygeoapi.config import get_config


def main():
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'django_.settings')
    django_app_path = Path(os.path.dirname(__file__))
    try:
        from django.core.management import execute_from_command_line
    except ImportError as exc:
        raise ImportError(
            'Couldn\'t import Django. Are you sure it\'s installed and '
            'available on your PYTHONPATH environment variable? Did you '
            'forget to activate a virtual environment?'
        ) from exc

    CONFIG = get_config()

    bind = f"{CONFIG['server']['bind']['host']}:{CONFIG['server']['bind']['port']}"  # noqa

    sys.argv = [str(django_app_path / 'django_app.py'), 'runserver', bind]

    sys.path.append(str(django_app_path))

    execute_from_command_line(sys.argv)


if __name__ == '__main__':
    main()



# =================================================================
#
# Authors: Alexander Pilz <a.pilz@52north.org>
#          Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2023 Alexander Pilz
# Copyright (c) 2023 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging
import time

from pygeoapi.process.base import BaseProcessor, ProcessorExecuteError

LOGGER = logging.getLogger(__name__)

#: Process metadata and description
PROCESS_METADATA = {
  'id': 'echo',
  'title': 'Echo Process',
  'description': 'Testable Echo process.',
  'version': '1.0.0',
  'jobControlOptions': [
    'async-execute',
    'sync-execute'
  ],
  'outputTransmission': [
    'value',
    'reference'
  ],
  'inputs': {
    'echoInput': {
      'title': 'Echo value',
      'description': 'Value to be echoed back.',
      'minOccurs': 1,
      'maxOccurs': 1,
      'schema': {
        'type': 'string',
        'enum': [
          'Echo',
          'Test',
          '42'
        ]
      }},
    'pause': {
      'title': 'Pause value',
      'description': 'Value to control the processing time.',
      'minOccurs': 1,
      'maxOccurs': 1,
      'schema': {
        'type': 'number',
        'enum': [
          5.5,
          10.25,
          42.0
        ]
      }
    }
  },
  'outputs': {
    'echoOutput': {
      'schema': {
        'type': 'string'
      }
    }
  },
  'links': [{
        'type': 'text/html',
        'rel': 'about',
        'title': 'information',
        'href': 'https://example.org/process',
        'hreflang': 'en-US'
    }],
  'example': {
    'inputs': {
      'echo': 'echoValue',
      'pause': 10.0
      }
  }
}


class EchoProcessor(BaseProcessor):
    """Echo Processor example"""
    def __init__(self, processor_def):
        """
        Initialize object

        :param processor_def: provider definition

        :returns: pygeoapi.process.echo.EchoProcessor
        """

        super().__init__(processor_def, PROCESS_METADATA)

    def execute(self, data, outputs=None):

        mimetype = 'application/json'

        echo = data.get('echoInput')
        pause = data.get('pause')

        if echo is None:
            raise ProcessorExecuteError(
                'Cannot run process without echo value')

        if not isinstance(echo, str):
            raise ProcessorExecuteError(
                'Cannot run process with echo not of type string')

        outputs = {
            'id': 'echoOutput',
            'value': echo
        }

        if pause is not None and isinstance(pause, float):
            time.sleep(pause)

        return mimetype, outputs

    def __repr__(self):
        return f'<EchoProcessor> {self.name}'



# =================================================================
#
# Authors: Timo Tuunanen <timo.tuunanen@rdvelho.com>
#
# Copyright (c) 2019 Timo Tuunanen
# Copyright (c) 2022 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

from datetime import datetime
import logging

from pymongo import MongoClient
from pymongo import GEOSPHERE
from pymongo import ASCENDING, DESCENDING
from pymongo.collection import ObjectId

from pygeoapi.crs import crs_transform
from pygeoapi.provider.base import BaseProvider, ProviderItemNotFoundError

LOGGER = logging.getLogger(__name__)


class MongoProvider(BaseProvider):
    """Generic provider for Mongodb.
    """

    def __init__(self, provider_def):
        """
        MongoProvider Class constructor

        :param provider_def: provider definitions from yml pygeoapi-config.
                             data,id_field, name set in parent class

        :returns: pygeoapi.provider.mongo.MongoProvider
        """
        # this is dummy value never used in case of Mongo.
        # Mongo id field is _id
        provider_def.setdefault('id_field', '_id')

        super().__init__(provider_def)

        LOGGER.info(f'Mongo source config: {self.data}')

        dbclient = MongoClient(self.data)
        self.featuredb = dbclient.get_default_database()
        self.collection = provider_def['collection']
        self.featuredb[self.collection].create_index([("geometry", GEOSPHERE)])
        self.get_fields()

    def get_fields(self):
        """
        Get provider field information (names, types)

        :returns: dict of fields
        """

        if not self._fields:
            pipeline = [
                {"$project": {"properties": 1}},
                {"$unwind": "$properties"},
                {"$group": {"_id": "$properties", "count": {"$sum": 1}}},
                {"$project": {"_id": 1}}
            ]

            result = list(self.featuredb[self.collection].aggregate(pipeline))

            # prepare a dictionary with fields
            # set the field type to 'string'.
            # by operating without a schema, mongo can query any data type.
            for i in result:
                for key in result[0]['_id'].keys():
                    self._fields[key] = {'type': 'string'}

        return self._fields

    def _get_feature_list(self, filterObj, sortList=[], skip=0, maxitems=1,
                          skip_geometry=False):
        featurecursor = self.featuredb[self.collection].find(filterObj)

        if sortList:
            featurecursor = featurecursor.sort(sortList)

        matchCount = self.featuredb[self.collection].count_documents(filterObj)
        featurecursor.skip(skip)
        featurecursor.limit(maxitems)
        featurelist = list(featurecursor)
        for item in featurelist:
            item['id'] = str(item.pop('_id'))
            if skip_geometry:
                item['geometry'] = None

        return featurelist, matchCount

    @crs_transform
    def query(self, offset=0, limit=10, resulttype='results',
              bbox=[], datetime_=None, properties=[], sortby=[],
              select_properties=[], skip_geometry=False, q=None, **kwargs):
        """
        query the provider

        :returns: dict of 0..n GeoJSON features
        """
        and_filter = []

        if len(bbox) == 4:
            x, y, w, h = map(float, bbox)
            and_filter.append(
                {'geometry': {'$geoWithin': {'$box': [[x, y], [w, h]]}}})

        # This parameter is not working yet!
        # gte is not sufficient to check date range
        if datetime_ is not None:
            assert isinstance(datetime_, datetime)
            and_filter.append({'properties.datetime': {'$gte': datetime_}})

        for prop in properties:
            and_filter.append({"properties."+prop[0]: {'$eq': prop[1]}})

        filterobj = {'$and': and_filter} if and_filter else {}

        sort_list = [("properties." + sort['property'],
                      ASCENDING if (sort['order'] == '+') else DESCENDING)
                     for sort in sortby]

        featurelist, matchcount = self._get_feature_list(
            filterobj, sortList=sort_list, skip=offset, maxitems=limit,
            skip_geometry=skip_geometry)

        if resulttype == 'hits':
            featurelist = []

        feature_collection = {
            'type': 'FeatureCollection',
            'features': featurelist,
            'numberMatched': matchcount,
            'numberReturned': len(featurelist)
        }

        return feature_collection

    @crs_transform
    def get(self, identifier, **kwargs):
        """
        query the provider by id

        :param identifier: feature id
        :returns: dict of single GeoJSON feature
        """
        featurelist, matchcount = self._get_feature_list(
                                    {'_id': ObjectId(identifier)})
        if featurelist:
            return featurelist[0]
        else:
            err = f'item {identifier} not found'
            LOGGER.error(err)
            raise ProviderItemNotFoundError(err)

    def create(self, new_feature):
        """Create a new feature
        """
        self.featuredb[self.collection].insert_one(new_feature)

    def update(self, identifier, updated_feature):
        """Updates an existing feature id with new_feature

        :param identifier: feature id
        :param new_feature: new GeoJSON feature dictionary
        """
        data = {k: v for k, v in updated_feature.items() if k != 'id'}
        self.featuredb[self.collection].update_one(
            {'_id': ObjectId(identifier)}, {"$set": data})

    def delete(self, identifier):
        """Deletes an existing feature

        :param identifier: feature id
        """
        self.featuredb[self.collection].delete_one(
            {'_id': ObjectId(identifier)})



# =================================================================
#
# Authors: Francesco Bartoli <xbartolone@gmail.com>
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#
# Copyright (c) 2025 Francesco Bartoli
# Copyright (c) 2022 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging
from http import HTTPStatus

from pygeoapi.models.provider.base import (
    TileMatrixSetEnum)

from pygeoapi.provider.base import (
    ProviderGenericError, ProviderItemNotFoundError)

LOGGER = logging.getLogger(__name__)


class BaseTileProvider:
    """generic Tile Provider ABC"""

    def __init__(self, provider_def):
        """
        Initialize object

        :param provider_def: provider definition

        :returns: pygeoapi.provider.tile.BaseTileProvider
        """

        self.name = provider_def['name']
        self.data = provider_def['data']
        self.format_type = provider_def['format']['name']
        self.mimetype = provider_def['format']['mimetype']
        self.options = provider_def.get('options')
        self.schemes = self.options.get('schemes', [])
        self.tile_type = None

    def get_layer(self):
        """
        Get provider layer name

        :returns: `string` of layer name
        """

    def get_fields(self):
        """
        Get provider field information (names, types)

        :returns: `dict` of fields
        """

        raise NotImplementedError()

    def get_tiling_schemes(self):
        """
        Get provider field information (names, types)

        :returns: `dict` of tiling schemes
        """

        raise NotImplementedError()

    def get_tiles_service(self, baseurl, servicepath, dirpath,
                          tile_type):
        """
        Gets tile service description

        :param baseurl: base URL of endpoint
        :param servicepath: base path of URL
        :param dirpath: directory basepath (equivalent of URL)
        :param tile_type: tile format type

        :returns: `dict` of file listing or `dict` of GeoJSON item or raw file
        """

        raise NotImplementedError()

    def get_tiles(self, layer, tileset, z, y, x, format_):
        """
        Gets tiles data

        :param layer: tile layer
        :param tileset: tile set
        :param z: z index
        :param y: y index
        :param x: x index
        :param format_: tile format type

        :returns: `binary` of the tile
        """

        raise NotImplementedError()

    def get_metadata(self):
        """
        Provide data/file metadata

        :returns: `dict` of metadata construct (format
                  determined by provider/standard)
        """

        raise NotImplementedError()

    def is_in_limits(self, tilematrixset, z, x, y):
        """
        Is within the limits of the tilematrixset

        :param z: tilematrix
        :param x: x
        :param y: y

        :returns: wether this tile is within the tile matrix
        set limits (Boolean)
        """

        try:
            x, y, z = map(int, (x, y, z))
        except ValueError:
            LOGGER.warning('Unable to cast tile index to integer')
            return False

        return all([
            x < tilematrixset.tileMatrices[z]['matrixWidth'],
            y < tilematrixset.tileMatrices[z]['matrixHeight'],
            self.options['zoom']['min'] <= z <= self.options['zoom']['max']
        ])

    def get_tilematrixset(self, tileMatrixSetId):
        """
        Get tilematrixset

        :param tileMatrixSetId: tilematrixsetid str

        :returns: tilematrixset enum object
        """

        enums = [e.value for e in TileMatrixSetEnum]
        enum = None

        try:
            for e in enums:
                if tileMatrixSetId == e.tileMatrixSet:
                    enum = e
            if not enum:
                raise ValueError('could not find this tilematrixset')
            return enum

        except ValueError as err:
            LOGGER.error(err)


class ProviderTileQueryError(ProviderGenericError):
    """provider tile query error"""
    default_msg = 'Tile not found'


class ProviderTileNotFoundError(ProviderItemNotFoundError):
    """provider tile not found error"""
    default_msg = 'Tile not found (check logs)'


class ProviderTilesetIdNotFoundError(ProviderTileQueryError):
    """provider tileset matrix query error"""
    default_msg = 'Tileset id not found'
    http_status_code = HTTPStatus.NOT_FOUND
    ogc_exception_code = 'NotFound'



# =================================================================
#
# Authors: Francesco Bartoli <xbartolone@gmail.com>
# Authors: Tom Kralidis <tomkralidis@gmail.com>
# Authors: Simon Seyock <simonseyock@gmail.com>
#
# Copyright (c) 2025 Francesco Bartoli
# Copyright (c) 2023 Tom Kralidis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

import logging

import requests
from urllib.parse import urlparse, urlencode

from pygeoapi.provider.tile import (ProviderTileNotFoundError,
                                    BaseTileProvider)
from pygeoapi.provider.base import (ProviderConnectionError,
                                    ProviderInvalidQueryError,
                                    ProviderGenericError)
from pygeoapi.models.provider.base import (
    TileMatrixSetEnum, TilesMetadataFormat, TileSetMetadata, LinkType)
from pygeoapi.util import is_url, url_join

LOGGER = logging.getLogger(__name__)


class WMTSFacadeProvider(BaseTileProvider):
    """WMTS Provider"""

    def __init__(self, provider_def):
        """
        Initialize object

        :param provider_def: provider definition

        :returns: pygeoapi.provider.WMTSFacadeProvider
        """

        if provider_def['format']['name'] not in ['png', 'jpeg']:
            raise RuntimeError("WMTS format has to be 'png' or 'jpeg'")

        super().__init__(provider_def)

        self.tile_type = 'raster'

    def __repr__(self):
        return f'<WMTSFacadeProvider> {self.data}'

    @property
    def service_url(self):
        return self._service_url

    @property
    def service_metadata_url(self):
        return self._service_metadata_url

    def get_tiling_schemes(self):

        tile_matrix_set_links_list = [
            TileMatrixSetEnum.WORLDCRS84QUAD.value,
            TileMatrixSetEnum.WEBMERCATORQUAD.value
        ]
        tile_matrix_set_links = [
            item for item in tile_matrix_set_links_list
            if item.tileMatrixSet == self.options['scheme']
        ]

        return tile_matrix_set_links

    def get_tiles_service(self, baseurl=None, servicepath=None,
                          dirpath=None, tile_type=None):
        """
        Gets service description

        :param baseurl: base URL of endpoint
        :param servicepath: base path of URL
        :param dirpath: directory basepath (equivalent of URL)
        :param tile_type: tile format type

        :returns: `dict` of item tile service
        """

        url = urlparse(self.data)
        baseurl = baseurl or f'{url.scheme}://{url.netloc}'
        tile_type = self.format_type
        basepath = url.path.split('/{z}/{x}/{y}')[0]
        servicepath = servicepath or f'{basepath}/tiles/{{tileMatrixSetId}}/{{tileMatrix}}/{{tileRow}}/{{tileCol}}?f={tile_type}'  # noqa

        if servicepath.startswith(baseurl):
            self._service_url = servicepath
        else:
            self._service_url = url_join(baseurl, servicepath)

        tile_matrix_set = self.service_url.split(
            '/{tileMatrix}/{tileRow}/{tileCol}')[0]
        self._service_metadata_url = url_join(tile_matrix_set, 'metadata')

        links = {
            'links': [
                {
                    'type': self.mimetype,
                    'rel': 'item',
                    'title': 'This collection as image tiles',
                    'href': self.service_url
                },
                {
                    'type': 'application/json',
                    'rel': 'describedby',
                    'title': 'Collection metadata in TileJSON format',
                    'href': f'{self.service_metadata_url}?f=json'
                }
            ]
        }
        return links

    def get_tiles(self, layer=None, tileset=None,
                  z=None, y=None, x=None, format_=None):
        """
        Gets tile

        :param layer: The layer name
        :param tileset: The tileset name
        :param z: z index
        :param y: y index
        :param x: x index
        :param format_: tile format

        :returns: an image tile
        """
        if is_url(self.data):
            params = {
                'service': 'WMTS',
                'request': 'GetTile',
                'version': '1.0.0',
                'format': self.mimetype,
                'layer': self.options['wmts_layer'],
                'tileMatrixSet': self.options['wmts_tile_matrix_set'],
                'tileMatrix': z,
                'tileRow': y,
                'tileCol': x,
                'style': self.options.get('wmts_style', '')
            }

            if '?' in self.data:
                request_url = '&'.join([self.data, urlencode(params)])
            else:
                request_url = '?'.join([self.data, urlencode(params)])

            LOGGER.debug(f'WMTS 1.0.0 request url: {request_url}')

            try:
                with requests.Session() as session:
                    resp = session.get(request_url)

                    if resp.status_code == 400:

                        tms = self.get_tilematrixset(tileset)

                        if (self.is_in_limits(tms, z, x, y)):
                            return None
                        raise ProviderTileNotFoundError

                    resp.raise_for_status()
                    return resp.content

            except requests.exceptions.RequestException as e:
                LOGGER.debug(e)
                if resp.status_code < 500:
                    raise ProviderInvalidQueryError  # Client is sending an invalid request # noqa
                raise ProviderGenericError  # Server error

        else:
            msg = f'Wrong data path configuration: {self.data}'
            LOGGER.error(msg)
            raise ProviderConnectionError(msg)

    def get_metadata(self, dataset, server_url, layer=None,
                     tileset=None, metadata_format=None, title=None,
                     description=None, keywords=None, **kwargs):
        """
        Gets tiles metadata

        :param dataset: dataset name
        :param server_url: server base url
        :param layer: mvt tile layer name
        :param tileset: mvt tileset name
        :param metadata_format: format for metadata,
                            enum TilesMetadataFormat
        :param title: title name
        :param description: description name
        :param keywords: keywords list

        :returns: `dict` of JSON metadata
        """

        if metadata_format.upper() == TilesMetadataFormat.JSON:
            return self.get_default_metadata(dataset, server_url, layer,
                                             tileset, title, description,
                                             keywords, **kwargs)
        elif metadata_format.upper() == TilesMetadataFormat.HTML:
            return self.get_html_metadata(dataset, server_url, layer,
                                          tileset, title, description,
                                          keywords, **kwargs)
        else:
            raise NotImplementedError(f"_{metadata_format.upper()}_ is not supported")  # noqa

    def get_html_metadata(self, dataset, server_url, layer, tileset,
                          title, description, keywords, **kwargs):

        service_url = url_join(
            server_url,
            f'collections/{dataset}/tiles/{tileset}/{{tileMatrix}}/{{tileRow}}/{{tileCol}}?f={self.format_type}')  # noqa
        metadata_url = url_join(
            server_url,
            f'collections/{dataset}/tiles/{tileset}/metadata')

        metadata = dict()
        metadata['id'] = dataset
        metadata['title'] = title
        metadata['tileset'] = tileset
        metadata['collections_path'] = service_url
        metadata['json_url'] = f'{metadata_url}?f=json'

        return metadata

    def get_default_metadata(self, dataset, server_url, layer, tileset,
                             title, description, keywords, **kwargs):

        service_url = url_join(
            server_url,
            f'collections/{dataset}/tiles/{tileset}/{{tileMatrix}}/{{tileRow}}/{{tileCol}}?f={self.format_type}')  # noqa

        content = {}
        tiling_schemes = self.get_tiling_schemes()
        # Default values
        tileMatrixSetURI = tiling_schemes[0].tileMatrixSetURI
        crs = tiling_schemes[0].crs
        # Checking the selected matrix in configured tiling_schemes
        for schema in tiling_schemes:
            if (schema.tileMatrixSet == tileset):
                crs = schema.crs
                tileMatrixSetURI = schema.tileMatrixSetURI

                tiling_scheme_url = url_join(
                    server_url, f'/TileMatrixSets/{schema.tileMatrixSet}')
                tiling_scheme_url_type = "application/json"
                tiling_scheme_url_title = f'{schema.tileMatrixSet} tile matrix set definition'  # noqa

                tiling_scheme = LinkType(href=tiling_scheme_url,
                                         rel="http://www.opengis.net/def/rel/ogc/1.0/tiling-scheme",  # noqa
                                         type_=tiling_scheme_url_type,
                                         title=tiling_scheme_url_title)

        if tiling_scheme is None:
            msg = f'Could not identify a valid tiling schema'  # noqa
            LOGGER.error(msg)
            raise ProviderConnectionError(msg)

        content = TileSetMetadata(title=title, description=description,
                                  keywords=keywords, crs=crs,
                                  tileMatrixSetURI=tileMatrixSetURI)

        links = []
        service_url_link_title = f'{tileset} raster tiles for {dataset}'
        service_url_link = LinkType(href=service_url, rel="item",
                                    type_=self.mimetype,
                                    title=service_url_link_title)

        links.append(tiling_scheme)
        links.append(service_url_link)

        content.links = links

        return content.model_dump(exclude_none=True)



# =================================================================
#
# Authors: Tom Kralidis <tomkralidis@gmail.com>
#          Ricardo Garcia Silva <ricardo.garcia.silva@geobeyond.it>
#
# Copyright (c) 2026 Tom Kralidis
# Copyright (c) 2023 Ricardo Garcia Silva
# Copyright (c) 2025 Angelos Tzotsos
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# =================================================================

__version__ = '0.23.4'

import click
try:
    # importlib.metadata is part of Python's standard library from 3.8
    from importlib.metadata import entry_points
except ImportError:
    from importlib_metadata import entry_points

from pygeoapi.asyncapi import asyncapi
from pygeoapi.config import config
from pygeoapi.openapi import openapi


def _find_plugins():
    """
    A decorator to find pygeoapi CLI plugins provided by third-party packages.

    pygeoapi plugins can hook into the pygeoapi CLI by providing their CLI
    functions and then using an entry_point named 'pygeoapi'.
    """

    def decorator(click_group):
        try:
            found_entrypoints = entry_points(group="pygeoapi")
        except TypeError:
            # earlier versions of importlib_metadata did not have the
            # `group` kwarg. More detail:
            #
            # https://github.com/geopython/pygeoapi/issues/1241#issuecomment-1536128897  # noqa: E501
            for group, entries in entry_points().items():
                if group == "pygeoapi":
                    found_entrypoints = entries
                    break
            else:
                found_entrypoints = []
        for entry_point in found_entrypoints:
            try:
                click_group.add_command(entry_point.load())
            except Exception as err:
                click.echo(err)
        return click_group

    return decorator


@click.group()
@click.version_option(version=__version__)
def cli():
    pass


@_find_plugins()
@cli.group()
def plugins():
    """Additional commands provided by third-party pygeoapi plugins"""
    pass


@cli.command()
@click.option('--flask', 'server', flag_value="flask", default=True)
@click.option('--starlette', 'server', flag_value="starlette")
@click.option('--django', 'server', flag_value="django")
@click.pass_context
def serve(ctx, server):
    """Run the server with different daemon type (--flask is the default)"""

    if server == "flask":
        from pygeoapi.flask_app import serve as serve_flask
        ctx.invoke(serve_flask)
    elif server == "starlette":
        from pygeoapi.starlette_app import serve as serve_starlette
        ctx.invoke(serve_starlette)
    elif server == "django":
        from pygeoapi.django_app import main as serve_django
        ctx.invoke(serve_django)
    else:
        raise click.ClickException('--flask/--starlette/--django is required')


cli.add_command(config)
cli.add_command(openapi)
cli.add_command(asyncapi)
